# coding: utf-8
"""
    RawData.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from SerialDataUtils import load_intensity_files, load_uv_array
from molass_legacy.KekLib.NumpyUtils import np_loadtxt
from molass_legacy._MOLASS.SerialSettings import get_setting, get_xray_picking
from MatrixData import simple_plot_3d

def get_row_index_impl(qvector, q):
    return min(len(qvector)-1, bisect_right(qvector, q))

class RawMatrix:
    def plot_3d(self, ax, alpha=1):
        simple_plot_3d(ax, self.data, x=self.vector, alpha=alpha)

class RawXray(RawMatrix):
    def __init__(self, in_folder):
        array, files = load_intensity_files(in_folder)
        assert len(array.shape) == 3
        array = self.apply_restriction(array)
        self.qvector = array[0,:,0]
        self.vector = self.qvector
        self.data = array[:,:,1].T
        self.error = array[:,:,2].T
        self.excluded = None

    def apply_restriction(self, array):
        xr_restrict_list = get_setting('xr_restrict_list')
        # print('xr_restrict_list=', xr_restrict_list)
        if xr_restrict_list is None:
            a_slice = slice(None)
            e_slice = slice(None)
            j0 = 0
        else:
            elution_info, angular_info = xr_restrict_list
            a_slice = slice(None) if angular_info is None else slice(angular_info.start, angular_info.stop)
            e_slice = slice(None) if elution_info is None else slice(elution_info.start, elution_info.stop)
            j0 = 0 if elution_info is None else elution_info.start
        self.j0 = j0
        return array[e_slice,a_slice,:].copy()

    def get_row_index(self, q=None):
        if q is None:
            q = get_xray_picking()
        return get_row_index_impl(self.qvector, q)

    def make_uv_proxy(self):
        i = self.get_row_index(0.02)
        return RawUvProxy(self.data[i,:], j0=self.j0)

    def exclude_elution(self, points):
        """
        ported
        from SerialData.exclude_intensities
        """
        from_ = None
        for i in points:
            if from_ is None:
                from_ = i
            else:
                if i > last + 1:
                    self.exclude_elution_impl(from_, last)
                    from_ = i

            last = i
        if from_ is not None:
            self.exclude_elution_impl(from_, last)
        self.excluded = points

    def exclude_elution_impl(self, from_, to_):
        """
        ported
        from SerialData.exclude_intensity
        """
        data = self.data

        debug = False
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            print('exclude_elution_impl: ', from_, to_)

            k = self.get_row_index(0.02)
            plt.push()
            plt.plot(data[k,:])
            plt.show()
            plt.pop()

        errr = self.error
        size = data.shape[1]
        if from_ == 0:
            j = to_ + 1
            for i in range(from_, j):
                if debug: print([i], 'extended')
                data[:,i] = data[:,j]
                errr[:,i] = errr[:,j]
        elif to_ == size - 1:
            j = from_ - 1
            for i in range(from_, size):
                if debug: print([i], 'extended')
                data[:,i] = data[:,j]
                errr[:,i] = errr[:,j]
        else:
            lower = from_ - 1
            upper = to_ + 1
            lower_data = data[:,lower]
            lower_errr = errr[:,lower]
            upper_data = data[:,upper]
            upper_errr = errr[:,upper]
            width = upper - lower
            for i in range(1, width):
                if debug: print([lower+i], 'interpolated with', lower, upper)
                w = i/width
                data[:,lower+i] = (1-w)*lower_data + w*upper_data
                errr[:,lower+i] = (1-w)*lower_errr + w*upper_errr

        if debug:
            k = self.get_row_index(0.02)
            plt.push()
            plt.plot(data[k,:])
            plt.show()
            plt.pop()

class RawUv(RawMatrix):
    def __init__(self, in_folder):
        array, vector, _ = load_uv_array(in_folder)
        assert array is not None
        array, vector = self.apply_restriction(array, vector)
        self.wvector = vector
        self.vector = vector
        self.data = array
        self.error = None

    def apply_restriction(self, array, vector):
        uv_restrict_list = get_setting('uv_restrict_list')
        # print('uv_restrict_list=', uv_restrict_list)
        if uv_restrict_list is None:
            w_slice = slice(None)
            e_slice = slice(None)
            j0 = 0
        else:
            elution_info, wavelen_info = uv_restrict_list
            w_slice = slice(None) if wavelen_info is None else slice(wavelen_info.start, wavelen_info.stop)
            e_slice = slice(None) if elution_info is None else slice(elution_info.start, elution_info.stop)
            j0 = 0 if elution_info is None else elution_info.start
        self.j0 = j0
        return array[w_slice, e_slice].copy(), vector[w_slice].copy()

    def get_row_index(self, w):
        return min(len(self.wvector)-1, bisect_right(self.wvector, w))

class RawUvProxy(RawUv):
    def __init__(self, ey, j0=0):
        import os
        import logging

        logger = logging.getLogger()
        logger.info("creating a RawUvProxy")

        this_dir = os.path.dirname( os.path.abspath( __file__ ) )
        ridge_data, _ = np_loadtxt(this_dir + '/../SerialAnalyzer/UV.dat')
        vector = ridge_data[:,0]
        curve_y = ridge_data[:,1]
        self.wvector = vector
        self.vector = vector
        self.data = np.dot(curve_y.reshape(len(curve_y),1), ey.reshape(1,len(ey)))
        self.error = None
        self.j0 = j0



"""

    DevelUtils.FolderWalkers.py

    Copyright (c) 2017-2023, SAXS Team, KEK-PF
"""
import sys
import os
import glob
import re

def do_a_folder(in_folder, cb):
    dat_files = glob.glob( in_folder + '/*.dat' )
    num_files = len( dat_files )
    if num_files < 90:
        return True

    txt_files = glob.glob( in_folder + '/*.txt' ) + glob.glob( in_folder + '/*.mtd' )
    num_txt_files = len( txt_files )
    if num_txt_files == 0:
        uv_folder = '/'.join( in_folder.split( '/' )[0:-1] ) + '/spectra'
        if os.path.exists( uv_folder ):
            # print( uv_folder )
            pass
        else:
            allow = False
            for name in ['20160628', '20191031', '_no_UV']:
                if in_folder.find(name) >= 0:
                    uv_folder = in_folder
                    allow = True
                    break
            if allow:
                pass
            else:
                print('--------------------- skipping', in_folder)
                return True
    else:
        uv_folder = in_folder

    return cb( in_folder, uv_folder=uv_folder, plot=False )

def serial_folder_walk(folder, cb, level=0, depth=3, reverse=False):
    for node in sorted(os.listdir(folder), reverse=reverse):
        path = '/'.join( [ folder, node ] )
        if not os.path.isdir( path ): continue
        # print(node)

        if node.find( 'autorg' ) >= 0: continue
        if node.find( 'analysis' ) >= 0: continue
        if node.find( 'averaged' ) >= 0: continue
        if node.find( '-stain' ) >= 0: continue
        if node.find( '_original' ) >= 0: continue
        if node.find( 'CirAve' ) >= 0: continue
        if node.find( '00_model' ) >= 0: continue
        if node.find( '_full' ) >= 0: continue
        if node.find( '-skip' ) >= 0: continue

        # print('path=', path)
        if path.find( '20170209_2' ) > 0: continue
        if path.find( '/JT' ) >= 0: continue
        if path.find( '20211105' ) > 0: continue

        ret = do_a_folder(path, cb)
        if not ret:
            return False
        if level < depth:
            ret = serial_folder_walk(path, cb, level=level+1, depth=depth, reverse=reverse)
            if not ret:
                return False
    return True



"""
    Automation.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import os
import logging
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkCustomWidgets import FileEntry

class AutomationDialog(Dialog):
    def __init__(self, parent, parent_dialog):
        self.parent_dialog = parent_dialog
        self.logger = logging.getLogger(__name__)
        Dialog.__init__(self, parent, title="Automation", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        label = Tk.Label(body_frame, text="Select your automation script")
        label.pack()
        filepath_frame = Tk.Frame(body_frame)
        filepath_frame.pack()
        path_label = Tk.Label(filepath_frame, text="Script File Path: ")
        path_label.pack(side=Tk.LEFT)
        self.scriptfile_path = Tk.StringVar()
        homedir = os.path.abspath(os.path.dirname(__file__) + r"/../..")
        self.scriptfile_path.set(os.path.join(homedir, r"test\automation.py"))
        file_entry = FileEntry(filepath_frame, textvariable=self.scriptfile_path, width=60, on_entry_cb=self.on_entry )
        file_entry.pack(side=Tk.LEFT)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Run", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def on_entry(self):
        pass

    def ok(self):
        self.run()
        self.cancel()

    def run(self):
        from molass_legacy.KekLib.ImportUtils import import_module_from_path
        from molass_legacy.KekLib.TkTester import TestClient

        file_path = self.scriptfile_path.get()
        mod = import_module_from_path("Runner", file_path)
        client = TestClient(self.parent_dialog, mod.run)
        print("Done")



"""
    Elution.Proportional.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.ExceptionTracebacker import log_exception

def decompose_by_binary_props(x, y, model, p1, p2, debug=False):
    # plot_info = xy_for_plot if debug else None
    plot_info = None
    params_array = model.guess_binary_peaks(x, y, p1, p2, debug=debug, plot_info=plot_info)
    target_props = np.array([p1, p2])

    def objective(p):
        areas = []
        cy_list = []
        for params in p.reshape(params_array.shape):
            cy = model(x, params)
            cy_list.append(cy)
            areas.append(np.sum(cy))
        try:
            ty = np.sum(cy_list, axis=0)
            props = np.array(areas)/np.sum(areas)
            fv = max(np.sum((ty - y)**2), np.sum((props - target_props)**2))
        except:
            # 
            fv = np.inf
        return fv

    ret = minimize(objective, params_array.flatten())
    params_array = ret.x.reshape(params_array.shape)
    return params_array

def fetch_component_by_prop(x, y, model, p, debug=False):
    params_array = model.guess_binary_peaks(x, y, p, 1-p, debug=False)
    return params_array[0]

def decompose_by_proportions_impl(x, y, model, target_props, debug=False):
    global xy_for_plot
    if debug:
        xy_for_plot = (x, y)
    sum_props = np.sum(target_props)
    if abs(sum_props - 1.0) > 1e-6:
        print("decompose_by_proportions: sum of target_props is not 1.0, they will be normalized.")
        target_props = target_props/sum_props

    resid_y = y.copy()
    params_list = []
    num_components = len(target_props)
    for j in range(num_components-1):
        p1 = target_props[j]
        p2 = np.sum(target_props[j+1:])
        total_p = p1 + p2
        p1 = p1/total_p
        p2 = p2/total_p
        print([j], p1, p2)
        try:
            if j < num_components-2:
                params = fetch_component_by_prop(x, resid_y, model, p1, debug=debug)
            else:
                params_array = decompose_by_binary_props(x, resid_y, model, p1, p2, debug=debug)
                params = params_array[0]
            cy = model(x, params)
            resid_y -= cy
            params_list.append(params)
        except:
            log_exception(None, "decompose_by_proportions: failed in %dth decopose_by_binary_props." % j)
        
    params_list.append(params_array[1])
    temp_params_array = np.array(params_list)

    topxv = temp_params_array[:,1]
    init_params = temp_params_array[:,[0,2,3,4]]

    def final_objective(p, title=None):
        areas = []
        cy_list = []
        for k, params in enumerate(p.reshape(init_params.shape)):
            params_ = np.array([params[0], topxv[k], *params[1:]])
            cy = model(x, params_)
            cy_list.append(cy)
            areas.append(np.sum(cy))
        try:
            ty = np.sum(cy_list, axis=0)
            props = np.array(areas)/np.sum(areas)
            fv = max(np.sum((ty - y)**2), np.sum((props - target_props)**2))
            if title is not None:
                print(title, fv)
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title(title)
                    ax.plot(x, y)
                    for cy in cy_list:
                        ax.plot(x, cy, ":")
                    ax.plot(x, ty, ":", color="red")
                    fig.tight_layout()
                    plt.show()
        except:
            #
            fv = np.inf
        return fv

    init_params_ = init_params.flatten()
    if debug:
        final_objective(init_params_, title="before final_objective")
    res = minimize(final_objective, init_params_, method="Nelder-Mead")
    if debug:
        final_objective(res.x, title="after final_objective")
    opt_params_array = res.x.reshape(init_params.shape)
    ret_paramd_list = []
    for k, params in enumerate(opt_params_array):
        params_ = np.array([params[0], topxv[k], *params[1:]])
        ret_paramd_list.append(params_)
    ret_params_array = np.array(ret_paramd_list)

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("decompose_by_proportions_impl")
            ax.plot(x, y)
            for params in ret_params_array:
                cy = model(x, params)
                px, py = model.get_peaktop_xy(x, params)
                ax.plot(x, cy, ":")
                ax.plot(px, py, "o", color="red")
            fig.tight_layout()
            plt.show()
    
    return ret_params_array

def decompose_by_proportions(fx, y, uv_y, model, target_props, xr_only=True, ask_return=False, debug=False):
    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("decompose_by_proportions: input", fontsize=20)
            ax1.plot(fx, uv_y)
            ax2.plot(fx, y)
            fig.tight_layout()
            plt.show()

    try:
        params_array_xr = decompose_by_proportions_impl(fx, y, model, target_props, debug=debug)
    except:
        log_exception(None, "decompose_by_proportions: failed to decompose XR", n=10)
        params_array_xr = None

    if xr_only:
        return params_array_xr
    
    try:
        params_array_uv = decompose_by_proportions_impl(fx, uv_y, model, target_props, debug=False)
    except:
        log_exception(None, "decompose_by_proportions: failed to decompose UV", n=10)
        params_array_uv = None

    assert not (params_array_xr is None and params_array_uv is None)
    if params_array_xr is None:
        params_array_xr = model.guess_from_the_other(fx, y, uv_y, params_array_uv)
    else:
        params_array_uv = model.guess_from_the_other(fx, uv_y, y, params_array_xr)

    if ask_return:
        dp_kwargs = dict(window_title="KPD minimization", 
                        button_spec=["Accept", "Cancel"], 
                        guide_message="If you accept, the model will be updated with the new parameters.")
        with plt.Dp(**dp_kwargs):
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("Change Num Components", fontsize=20)
            ax1.plot(fx, uv_y)
            cy_list = []
            for params in params_array_uv:
                cy = model(fx, params)
                cy_list.append(cy)
                ax1.plot(fx, cy, ":")
            ty = np.sum(cy_list, axis=0)
            ax1.plot(fx, ty, ":", color="red")

            ax2.plot(fx, y)
            cy_list = []
            for params in params_array_xr:
                cy = model(fx, params)
                cy_list.append(cy)
                ax2.plot(fx, cy, ":")
            ty = np.sum(cy_list, axis=0)
            ax2.plot(fx, ty, ":", color="red")
            fig.tight_layout()
            ret = plt.show()
    if not ask_return or ret:
        return params_array_xr, params_array_uv
    else:
        return None


"""
    Optimizer.SecConformance.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from tksheet import Sheet
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from DataUtils import get_in_folder
from molass_legacy.KekLib.BasicUtils import ordinal_str, Struct
from SecTheory.RetensionTime import compute_retention_time

class SecConformance(Dialog):
    def __init__(self, parent, parent_dialog, state_canvas):
        self.parent = parent
        self.parent_dialog = parent_dialog
        self.state_canvas = state_canvas
        self.fullopt = state_canvas.fullopt
        self.x_array = state_canvas.demo_info[1]
        self.curr_index = state_canvas.get_curr_index()
        self.params = self.x_array[self.curr_index]
        self.n = self.fullopt.n_components
        self.separate_params = self.fullopt.split_params_simple(self.params)
        self.seccol_params = self.separate_params[-1]
        self.separate_optimization = Tk.IntVar()
        self.separate_optimization.set(0)
        self.sheet_frame = None
        self.sheet = None
        Dialog.__init__(self, parent, "SEC Conformance", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):

        figure_frame = Tk.Frame(body_frame)
        figure_frame.pack(side=Tk.LEFT)
        self.right_frame = Tk.Frame(body_frame)
        self.right_frame.pack(side=Tk.LEFT, padx=20)

        cframe = Tk.Frame(figure_frame)
        cframe.pack()
        tframe = Tk.Frame(figure_frame)
        tframe.pack(fill=Tk.X, padx=20)
        tframe_left = Tk.Frame(tframe)
        tframe_left.pack(side=Tk.LEFT)

        self.fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 6))
        axt = ax1.twinx()
        self.axes = (ax1, ax2, axt)
        self.show_sec_conformance()

        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X)
        for j in range(2):
            box.columnconfigure(j, weight=1)

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.grid(row=0, column=0, pady=5)

        option_frame = Tk.Frame(box)
        option_frame.grid(row=0, column=1)
        cb = Tk.Checkbutton(option_frame, text="separate optimization", variable=self.separate_optimization)
        cb.pack(side=Tk.LEFT)
        w = Tk.Button(option_frame, text="Refresh", width=10, command=self.show_sec_conformance)
        w.pack(side=Tk.LEFT)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show_sec_conformance(self):
        rgs = self.separate_params[2]
        trs = self.separate_params[0][:,1]

        if self.separate_optimization.get():
            result = self.fullopt.params_type.estimate_conformance_params(rgs, trs)
            model_trs = compute_retention_time(result.x, rgs)
        else:
            seccol_params = self.separate_params[-1]
            model_trs = compute_retention_time(seccol_params, rgs)
            result = self.make_pseudo_result(seccol_params, trs, model_trs)

        fig = self.fig
        for ax in self.axes:
            ax.cla()
        ax1, ax2, axt = self.axes
        axt.grid(False)

        in_folder = get_in_folder()
        job_info = self.state_canvas.dialog.get_job_info()
        job_name = job_info[0]

        fig.suptitle("Conformance with Stochastic Theory of SEC on %s job %s at %s state" % (in_folder, job_name, ordinal_str(self.curr_index)), fontsize=20)
        ax1.set_title("Xray Decomposition", fontsize=16)

        axis_info = (self.fig, (None, ax1, None, axt))
        self.state_canvas.fullopt.objective_func(self.params, debug=True, axis_info=axis_info)

        ax2.set_title("Single Pore Model Conformance: Score=%.3g" % result.fun, fontsize=16)
        ax2.set_xlabel("Retention Time (Eno)")
        ax2.set_ylabel("Rg")
        ax2.plot(model_trs, rgs, "o-", label="stochastic model")
        ax2.plot(trs, rgs, "o-",label="currently optimized")
        ax2.legend()

        fig.tight_layout()
        fig.subplots_adjust(top=0.85)

        self.make_data_sheet(rgs, trs, result, model_trs)

    def make_data_sheet(self, rgs, trs, result, model_trs):
        if self.sheet_frame is not None:
            self.sheet_frame.destroy()
        if self.sheet is not None:
            self.sheet.destroy()

        self.sheet_frame = Tk.Frame(self.right_frame)
        self.sheet_frame.pack()
        sheet_frame = self.sheet_frame
        sheet_title = Tk.Label(sheet_frame, text="Conformance Data Sheet", font=(None, 20))
        sheet_title.pack(pady=10)
        sheet_body = Tk.Frame(sheet_frame)
        sheet_body.pack(pady=10)

        data_list = self.get_sheet_data_list(rgs, trs, result, model_trs)

        num_rows = len(rgs) +12
        num_columns = 3
        column_width = 90
        width = column_width*num_columns + 60
        height = int(22*num_rows) + 60
        self.sheet = Sheet(sheet_body, width=width, height=height, data=data_list, show_selected_cells_border=False, column_width=column_width)
        self.sheet.pack()
        self.sheet.enable_bindings()

    def get_sheet_data_list(self, rgs, trs, result, model_trs):
        colnames = ["Rg", "Model tR", "Solution tR"]
        data_list = [colnames]
        for rg, model_tr, tr in zip(rgs, model_trs, trs):
            data_list.append(["%.3g" % v for v in [rg, model_tr, tr]])

        data_list.append([])
        t0, K, rp, m = result.x
        data_list.append(["rp (pore size)", "%.3g" % rp])
        data_list.append(["t0", "%.3g" % t0])
        data_list.append(["K", "%.3g" % K])
        data_list.append(["m", "%.3g" % m])

        data_list.append([])
        data_list.append(["fit success", str(result.success)])
        data_list.append(["fit status", str(result.status)])
        data_list.append(["fit message", result.message])

        data_list.append([])
        data_list.append(["score", "%.3g" % result.fun])

        return data_list

    def make_pseudo_result(self, seccol_params, trs, model_trs):
        fun = np.sqrt(np.average((model_trs - trs)**2))
        return Struct(x=seccol_params, success=True, status=0, message="Global", fun=fun)



"""
    ATSAS.AtsasVersion.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import os
import glob

def get_atsas_bin_path():
    path = os.path.join(os.environ["ATSAS"], "bin")
    path = path.replace('"', '')
    return path

def get_version():
    # implement another one separately because
    # importing from (or unifying) .AtsasUtils causes a peculiar problem
    from OurImporter import import_module_from_path
    packages_pattern = get_atsas_bin_path() + r'\python*\site-packages'
    path_list = glob.glob(packages_pattern)
    if len(path_list) > 0:
        packages_path = path_list[0]
        file_path = os.path.join(packages_path, "atsas.py")
        atsas = import_module_from_path('atsas', file_path)
        ver = atsas.ATSAS_VERSION
    else:
        ver = 'Unkown'
    return ver

def atsas_version_for_publication():
    return "ATSAS " + get_version()



# coding: utf-8
"""
    SvdViewer.py

    Copyright (c) 2018-2022, SAXS Team, KEK-PF
"""
import copy
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkUtils import is_low_resolution
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy._MOLASS.SerialSettings import get_setting
from ScatteringBaseCorrector import ScatteringBaseCorrector
from ScatteringViewUtils import compute_baselines, draw_3d_scattering
from molass_legacy.UV.AbsorbanceViewUtils import draw_3d

DEBUG = False

TITLE_FONTSIZE = 16
LABLE_FONTSIZE = 16
TICKLABEL_SIZE = 7
MAX_NUM_SVALUES = 10

ADD_TOOLBAR = False
if ADD_TOOLBAR:
    from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar


class SvdViewer(Dialog):
    def __init__(self, parent, title, sd, mapper):
        self.parent = parent
        self.title_ = title
        self.sd = sd
        self.mapper = mapper

    def show(self):
        self.parent.config(cursor='wait')
        Dialog.__init__(self, self.parent, self.title_)

    def body(self, body_frame):  # overrides parent class method
        tk_set_icon_portable(self)

        self.vframe = vframe = Tk.Frame(body_frame)
        vframe.pack()

        absorbance = self.sd.absorbance
        data = absorbance.get_corrected_data()
        ecurve = self.mapper.a_curve.y
        curves = [data[:, int(info[1] + 0.5)] for info in self.mapper.a_curve.peak_info]
        self.uv_frame = SvdFrame(vframe, self.sd.lvector, data, [ecurve], curves, "UV Absorbance",
                                 absorbance=absorbance, toggle_cb=self.show_the_other, toggle_text="Show Xray",
                                 dialog=self)
        self.uv_frame.grid(row=0, column=0)
        self.uv_frame.grid_forget()

        data = self.sd.intensity_array[:, :, 1].T
        ecurve = self.mapper.x_curve.y
        curves = [data[:, int(info[1] + 0.5)] for info in self.mapper.x_curve.peak_info]
        self.xray_frame = SvdFrame(vframe, self.sd.qvector, data, [ecurve], curves, "Xray Scattering",
                                   serial_data=self.sd, mapper=self.mapper,
                                   toggle_cb=self.show_the_other, toggle_text="Show UV",
                                   dialog=self)
        self.xray_frame.grid(row=0, column=0)
        self.xray_frame.grid_forget()

        self.bframe = bframe = Tk.Frame(body_frame)
        bframe.pack()

        self.current_data = 1
        # self.toggle_btn = Tk.Button( bframe, text="", command=self.show_the_other )
        # self.toggle_btn.grid( row=0, column=0 )
        self.show_the_other()
        self.parent.config(cursor='')

    def show_the_other(self):
        self.current_data = 1 - self.current_data
        if self.current_data == 0:
            self.uv_frame.grid_forget()
            self.xray_frame.grid(row=0, column=0)
            self.xray_frame.draw()
            # self.toggle_btn.config(text="Show UV")
        else:
            self.xray_frame.grid_forget()
            self.uv_frame.grid(row=0, column=0)
            self.uv_frame.draw()
            # self.toggle_btn.config(text="Show Xray")


class SvdFrame(Tk.Frame):
    def __init__(self, parent, vector, array, ecurves, ucurves, data_name,
                 serial_data=None, mapper=None, absorbance=None,
                 toggle_cb=None, toggle_text=None, dialog=None):
        self.parent = parent
        self.vector = vector
        self.array = array
        self.ecurves = ecurves
        self.ucurves = ucurves
        self.data_name = data_name
        self.serial_data = serial_data
        self.mapper = mapper
        self.absorbance = absorbance
        self.toggle_cb = toggle_cb
        self.toggle_text = toggle_text
        self.dialog = dialog

        U, s, VT = np.linalg.svd(array)
        self.U = U
        self.s = s
        self.V = VT.T
        Tk.Frame.__init__(self, self.parent)
        self.build_body()

    def build_body(self):  # overrides parent class method
        tk_set_icon_portable(self)

        body_frame = Tk.Frame(self)
        body_frame.pack()

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        ctlframe = Tk.Frame(body_frame)
        ctlframe.pack(fill=Tk.X)

        if is_low_resolution():
            figsize = (17, 7)
        else:
            figsize = (21, 10)

        self.fig = fig = plt.figure(figsize=figsize)

        in_folder = get_setting('in_folder')
        fig.suptitle(self.data_name + ' from ' + in_folder, fontsize=TITLE_FONTSIZE)

        gs = gridspec.GridSpec(2, 7)
        ax0 = self.fig.add_subplot(gs[:, 0:3], projection='3d')
        ax1 = self.fig.add_subplot(gs[0, 3:5])
        ax2 = self.fig.add_subplot(gs[1, 3:5])
        ax3 = self.fig.add_subplot(gs[:, 5:7])

        self.axes = [ax0, ax1, ax2, ax3]
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas.draw()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        if ADD_TOOLBAR:
            self.toolbar = NavigationToolbar(self.mpl_canvas, cframe)
            self.toolbar.update()

        ctlframe.columnconfigure(0, weight=1)
        ctlframe.columnconfigure(1, weight=1)

        toggle_frame = Tk.Frame(ctlframe)
        toggle_frame.grid(row=0, column=0)

        toggble_btn = Tk.Button(toggle_frame, text=self.toggle_text, command=self.toggle_cb)
        toggble_btn.pack()

        control_panel = Tk.Frame(ctlframe)
        control_panel.grid(row=0, column=1, sticky=Tk.E, padx=40)

        spinbox_frame = Tk.Frame(control_panel)
        spinbox_frame.grid(row=0, column=0, padx=10)

        spinbox_label = Tk.Label(spinbox_frame, text="Number of components to show: ")
        spinbox_label.grid(row=0, column=0)

        self.num_components = Tk.IntVar()
        self.num_components.set(2)

        self.spinbox = Tk.Spinbox(spinbox_frame, textvariable=self.num_components,
                                  from_=1, to=MAX_NUM_SVALUES, increment=1,
                                  justify=Tk.CENTER, width=6)
        self.spinbox.grid(row=0, column=1)

        self.uv_sign = 1

        self.reverse_button = Tk.Button(control_panel, text='Reverse', command=self.reverse)
        self.reverse_button.grid(row=0, column=1, padx=10)

        space = Tk.Frame(control_panel, width=400)
        space.grid(row=0, column=2)

        self.save_button = Tk.Button(control_panel, text='Results Saver', command=self.show_save_dialog)
        self.save_button.grid(row=0, column=3, padx=10)

        self.draw3d()
        self.fig.subplots_adjust(top=0.9, bottom=0.06, left=0.02, right=0.98, wspace=0.32, hspace=0.2)

        self.num_components.trace('w', self.num_components_tracer)

    def num_components_tracer(self, *args):
        self.draw()

    def reverse(self):
        if self.reverse_button.config('relief')[-1] == 'sunken':
            self.reverse_button.config(relief="raised")
            self.uv_sign = 1
        else:
            self.reverse_button.config(relief="sunken")
            self.uv_sign = -1

        self.draw()

    def draw3d(self):
        ax0 = self.axes[0]
        title = self.data_name + ' in 3D'

        if self.absorbance is None:
            serial_data = self.serial_data
            assert serial_data is not None
            assert self.mapper is not None

            data = serial_data.intensity_array
            qvector = serial_data.qvector
            slice_ = serial_data.xray_slice
            index = (slice_.start + slice_.stop) // 2
            xray_curve_y = serial_data.xray_curve.y
            opt_params = self.mapper.opt_params

            data_copy = copy.deepcopy(data)
            corrector = ScatteringBaseCorrector(
                serial_data.jvector,
                serial_data.qvector,
                data_copy,
                curve=serial_data.xray_curve,
                affine_info=self.mapper.get_affine_info(),
                inty_curve_y=xray_curve_y,
                baseline_opt=opt_params['xray_baseline_opt'],
                baseline_type=opt_params['xray_baseline_type'],
                need_adjustment=opt_params['xray_baseline_adjust'] == 1,
                parent=self)

            vsa_base_list, all_base_list = compute_baselines(qvector, index, corrector)

            draw_3d_scattering(ax0, data, qvector, index, xray_curve_y,
                               title, vsa_base_list, all_base_list
                               )
        else:
            absorbance = self.absorbance
            # TODO: consider unifying these params with those of AbsorbanceViewer
            wvlen_lower = 245
            wvlen_upper = 450
            i_slice = slice(0, len(absorbance.wl_vector))
            draw_3d(ax0, absorbance, wvlen_lower, wvlen_upper, i_slice,
                    title=title, title_fontsize=TITLE_FONTSIZE, low_percentile=False)

    def draw(self):
        num_components = self.num_components.get()
        self.draw_U(num_components)
        self.draw_V(num_components)
        self.draw_s(num_components)

        for ax in self.axes:
            if ax is None:
                continue
            ax.tick_params(labelsize=LABLE_FONTSIZE)

        self.mpl_canvas.draw()

    def draw_U(self, num_components):
        ax1 = self.axes[1]
        ax1.cla()
        ax1.set_title('Major U Spectra', fontsize=TITLE_FONTSIZE)
        # ax1.tick_params(axis='y', labelsize=TICKLABEL_SIZE)         # does not work

        for i in range(num_components):
            ax1.plot(self.vector, self.U[:, i] * self.uv_sign, label='U%d' % i, linewidth=5)

        ax1.legend(fontsize=LABLE_FONTSIZE)

    def draw_V(self, num_components):
        ax2 = self.axes[2]
        ax2.cla()
        ax2.set_title('Major V Spectra', fontsize=TITLE_FONTSIZE)
        # ax2.tick_params(axis='y', labelsize=TICKLABEL_SIZE)         # does not work

        for i in range(num_components):
            ax2.plot(self.V[:, i] * self.uv_sign, 'o', label='V%d' % i)

        ax2.legend(fontsize=LABLE_FONTSIZE)

    def draw_s(self, num_components):
        ax = self.axes[3]
        ax.cla()
        ax.set_title('Major Singular Values', fontsize=TITLE_FONTSIZE)

        num_see_values = self.num_components.get()
        num_show_values = max(6, min(MAX_NUM_SVALUES, num_see_values + 2))

        ax.plot(self.s[0:num_show_values], ':', color='gray')

        for i in range(num_show_values):
            ax.plot(i, self.s[i], 'o', markersize=12)

    def show_save_dialog(self):
        from SvdDataSaver import SvdDataSaverDialog
        dialog = SvdDataSaverDialog(self.dialog, self.data_name,
                        svd_results=[self.vector, self.U, self.s, self.V],
                        num_components=self.num_components.get(),
                        location='lower right')
        dialog.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Generate RDF from PDB
# Copyright (C) 2011 Stas Bevc, stas.bevc@cmm.ki.si
# www.sicmm.org

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

from multiprocessing import Pool, cpu_count
from sys import argv
from os import listdir
from math import sqrt, ceil, floor, pow, pi
import time


# generate radial distribution function from
# particle positions stored in PDB files
def makeRDF(pdbdir, pdbs, side, numbins, cm, numAT=0):
    
    maxbin = numbins # number of bins
    sideh = side[1]/2.0 # take y dir
    sidehx = side[0]/2.0
    sidehy = side[1]/2.0
    sidehz = side[2]/2.0
    dr = float(sideh/maxbin) # bin width
    hist = [0]*(maxbin+1) # initialize list
    count = 0
    cnt = 0
    
    # list of pdbs is now an input
    #pdbs = listdir(pdbdir) # list of files
    nstep = len(pdbs)
    
    cutsq = sideh*sideh
    
    mO = 1
    mH = 0.063002362588597
    mMol = 2*mH + mO
    
    # loop over pdb files
    for pdb in pdbs:
        
        #if not pdb.endswith(".pdb"):
        #    nstep -= 1
        #    continue # skip other files
            
        count += 1
        print "Reading file ... "+pdb+" ("+ str(count) +"/"+ str(len(pdbs)) +")"
        
        # read atom coordinates from PDB
        atoms = []
        cmm = [0,0,0] # center of mass of molecule
        atc = 1 # atom counter
        lines = open(pdbdir+"/"+pdb)
        for line in lines:
            if line[0:4] != "ATOM":
                continue # skip other lines
            coords = map(float, (line[31:54]).split())
            label = line[13:16] # atom label
            atype = line[77:78] # atom type
            
            if label == "NA " or label == "CL ":
                continue # skip other atoms
            
            
            if cm == True: # calculate center of mass
                if label == "OH ": # O
                    for i in range(3):
                        cmm[i] += mO * coords[i]
                if label == "1HH" or label == "2HH": # H1, H2
                    for i in range(3):
                        cmm[i] += mH * coords[i]
                if atc < numAT:
                    atc += 1
                else:
                    atc = 1
                    for i in range(3):
                        cmm[i] /= mMol
                    
                    # fold coordinates
                    for i in range(3):
                        tmp = floor(cmm[i] * (1.0/side[i]))
                        cmm[i] -= tmp * side[i]
                    
                    atoms.append((cmm[0],cmm[1],cmm[2]))
                    cmm = [0,0,0]
            else: # no cm calculation
                atoms.append((coords[0], coords[1], coords[2]))
        
        # loop over particle pairs
        npart = len(atoms)
        print " looping over particle pairs (" +str(npart)+ "^2) ... "
        for i in range(npart):
            
            xi = (atoms[i])[0]
            yi = (atoms[i])[1]
            zi = (atoms[i])[2]
            
            for j in range(i+1, npart):
                xx = xi - (atoms[j])[0]
                yy = yi - (atoms[j])[1]
                zz = zi - (atoms[j])[2]
                
                # minimum image
                if (xx < -sidehx):   xx = xx + side[0]
                elif (xx > sidehx):    xx = xx - side[0]
                if (yy < -sidehy):   yy = yy + side[1]
                elif (yy > sidehy):    yy = yy - side[1]
                if (zz < -sidehz):   zz = zz + side[2]
                elif (zz > sidehz):    zz = zz - side[2]
                
                # square distance between i and j
                rd  = xx * xx + yy * yy + zz * zz
                
                if rd < cutsq:
                    rij = sqrt(rd)
                    bin = int(ceil(rij/dr)) # determine in which bin the distance falls
                    if (bin <= maxbin):
                        hist[bin] += 1
                cnt += 1
    
    return (hist, cnt)
    
    
def normalize(hist, side, cnt):
    
    print "Normalizing ... "
    rdf = {}
    maxbin = (len(hist)-1)
    dr = float((side[1]/2.0)/maxbin) # bin width
    vol = side[0] * side[1] * side[2]  #pow(side, 3.0)
    
    for i in range(1, maxbin+1):
        rrr = (i - 0.5) * dr
        bin_vol = (4.0/3.0) * pi * (pow(rrr+(dr/2.0),3.0) - pow(rrr-(dr/2.0), 3.0))
        val = ((hist[i] * vol) / (cnt * bin_vol))
        rdf.update({rrr:val})
        
    #print "Normalizing ... "
    #rdf = {}
    #maxbin = (len(hist)-1)
    #dr = (side/2.0)/maxbin # bin width
    #phi = npart/pow(side, 3.0) # number density (N/V)
    #norm = 2.0 * pi * dr * phi * nstep * npart
    
    #print npart, nstep
    
    #for i in range(1, maxbin+1):
        #rrr = (i - 0.5) * dr
        #val = hist[i]/ norm / ((rrr * rrr) + (dr * dr) / 12.0)
        #rdf.update({rrr:val})
    
    return rdf




#-------------------------------------------------------------------#

if __name__ == '__main__':
    cpucount = cpu_count()
    pool = Pool(processes=cpucount)  # start worker processes

    # RDF setup
    pdbdir = argv[1] # directory with PDB files (eg. "./pdb-ex-10k/")
    boxsize = (51.16866708780796, 12.79216677195199, 12.79216677195199)
    numbins = 300 # number of bins
    outfile = pdbdir + "rdf.out"
    
    # calculate RDF from center of mass of molecule
    numAT = 3 # number of atoms in molecule
    cm = False # default to false
    #if len(argv) == 3:
    #    cm = argv[2]
    
    # prepare lists of pdb files
    files = listdir(pdbdir) # list of files
    nfiles = 0 # number of total pdb files
    pdbs = [] # list of pdb files
    for f in files:
        if not f.endswith(".pdb"):
            continue # skip other files
        nfiles += 1
        pdbs.append(f)
    
    if cpucount > nfiles:
        cpucount = nfiles
    
    files_oncpu = int(floor(nfiles / cpucount))
    dist_pdbs = [] # pdbs to distribute
    p = 0
    for i in range (cpucount):
        tmp = []
        for j in range(files_oncpu):
            tmp.append(pdbs[p])
            p += 1
        dist_pdbs.append(tmp)
    
    # distribute remaining files
    rem = nfiles - (files_oncpu * cpucount)
    if (rem > 0):
        for i in range(rem):
            dist_pdbs[i].append(pdbs[(files_oncpu * cpucount)+i])
    
    # distribute lists to pools
    results = [ pool.apply_async(makeRDF, [pdbdir,dist_pdbs[i],boxsize,numbins,cm,numAT])
                for i in range(len(dist_pdbs)) ]
    
    # sum up results
    hist = [0]*(numbins+1) # initialize list
    cnt = 0;
    for i in range(len(dist_pdbs)):
        result = results[i].get()
        cnt += result[1]
        for j in range(numbins+1):
            hist[j] += result[0][j]
    
    #print hist

    rdf = normalize(hist, boxsize, cnt)
    
    print "Writing output file ... " +outfile+ " (" + time.strftime("%c") +")"
    outfile = open(outfile, "w")
    for r in sorted(rdf.iterkeys()): # sort before writing into file
        outfile.write("%15.8g %15.8g\n"%(r, rdf[r]))
    outfile.close()






"""
    PyChemLib.py

    functions borrowed from
    https://pythoninchemistry.org/sim_and_scat/intro

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.constants import Boltzmann
mass_of_argon = 39.948 # amu

def lj_force(r, epsilon, sigma):
    """
    Implementation of the Lennard-Jones potential 
    to calculate the force of the interaction.
    
    Parameters
    ----------
    r: float
        Distance between two particles (Å)
    epsilon: float 
        Potential energy at the equilibrium bond 
        length (eV)
    sigma: float 
        Distance at which the potential energy is 
        zero (Å)
    
    Returns
    -------
    float
        Force of the van der Waals interaction (eV/Å)
    """
    return 48 * epsilon * np.power(
        sigma, 12) / np.power(
        r, 13) - 24 * epsilon * np.power(
        sigma, 6) / np.power(r, 7)

def demo_lj_force():
    r = np.linspace(3.5, 8, 100)
    plt.plot(r, lj_force(r, 0.0103, 3.4))
    plt.xlabel(r'$r$/Å')
    plt.ylabel(r'$f$/eVÅ$^{-1}$')
    plt.show()

def demo_lj_force2():
    r = np.linspace(5.9, 15, 100)
    plt.plot(r, lj_force(r, 0.1, 5.99))
    plt.xlabel(r'$r$/Å')
    plt.ylabel(r'$f$/eVÅ$^{-1}$')
    plt.show()

def init_velocity(T, number_of_particles, mass):
    """
    Initialise the velocities for a series of 
    particles.
    
    Parameters
    ----------
    T: float
        Temperature of the system at 
        initialisation (K)
    number_of_particles: int
        Number of particles in the system
    
    Returns
    -------
    ndarray of floats
        Initial velocities for a series of 
        particles (eVs/Åamu)
    """
    R = np.random.rand(number_of_particles) - 0.5
    return R * np.sqrt(Boltzmann * T / (
        mass * 1.602e-19))

def get_accelerations(positions, mass):
    """
    Calculate the acceleration on each 
    particle as a  result of each other 
    particle. 
    N.B. We use the Python convention of
    numbering from 0.
    
    Parameters
    ----------
    positions: ndarray of floats
        The positions, in a single dimension, 
        for all of the particles (Å)
        
    Returns
    -------
    ndarray of floats
        The acceleration on each particle (eV/Åamu)
    """
    accel_x = np.zeros((positions.size, positions.size))
    for i in range(0, positions.size - 1):
        for j in range(i + 1, positions.size):
            r_x = positions[j] - positions[i]
            rmag = np.sqrt(r_x * r_x)
            force_scalar = lj_force(rmag, 0.0103, 3.4)
            force_x = force_scalar * r_x / rmag
            accel_x[i, j] = force_x / mass #eV Å-1 amu-1
            # appling Newton's third law
            accel_x[j, i] = - force_x / mass
    return np.sum(accel_x, axis=0)

def demo_accel():
    mass_of_argon = 39.948 # amu
    accel = get_accelerations(np.array([1, 5, 10]), mass_of_argon)
    print('Acceleration on particle 0 = {:.3e} eV/Åamu'.format(accel[0]))
    print('Acceleration on particle 1 = {:.3e} eV/Åamu'.format(accel[1]))
    print('Acceleration on particle 2 = {:.3e} eV/Åamu'.format(accel[2]))

def update_pos(x, v, a, dt):
    """
    Update the particle positions.
    
    Parameters
    ----------
    x: ndarray of floats
        The positions of the particles in a 
        single dimension
    v: ndarray of floats
        The velocities of the particles in a 
        single dimension
    a: ndarray of floats
        The accelerations of the particles in a 
        single dimension
    dt: float
        The timestep length
    
    Returns
    -------
    ndarray of floats:
        New positions of the particles in a single 
        dimension
    """
    return x + v * dt + 0.5 * a * dt * dt

def update_velo(v, a, a1, dt):
    """
    Update the particle velocities.
    
    Parameters
    ----------
    v: ndarray of floats
        The velocities of the particles in a 
        single dimension (eVs/Åamu)
    a: ndarray of floats
        The accelerations of the particles in a 
        single dimension at the previous 
        timestep (eV/Åamu)
    a1: ndarray of floats
        The accelerations of the particles in a
        single dimension at the current 
        timestep (eV/Åamu)
    dt: float
        The timestep length
    
    Returns
    -------
    ndarray of floats:
        New velocities of the particles in a
        single dimension (eVs/Åamu)
    """
    return v + 0.5 * (a + a1) * dt

def run_md(dt, number_of_steps, initial_temp, x, mass):
    """
    Run a MD simulation.
    
    Parameters
    ----------
    dt: float
        The timestep length (s)
    number_of_steps: int
        Number of iterations in the simulation
    initial_temp: float
        Temperature of the system at 
        initialisation (K)
    x: ndarray of floats
        The initial positions of the particles in a 
        single dimension (Å)
        
    Returns
    -------
    ndarray of floats
        The positions for all of the particles 
        throughout the simulation (Å)
    """
    positions = np.zeros((number_of_steps, 3))
    v = init_velocity(initial_temp, 3, mass)
    a = get_accelerations(x, mass)
    for i in range(number_of_steps):
        x = update_pos(x, v, a, dt)
        a1 = get_accelerations(x, mass)
        v = update_velo(v, a, a1, dt)
        a = np.array(a1)
        positions[i, :] = x
    return positions

def demo_run_md():
    x = np.array([1, 5, 10])
    sim_pos = run_md(0.1, 10000, 300, x, mass_of_argon)
        
    # %matplotlib inline
    for i in range(sim_pos.shape[1]):
        plt.plot(sim_pos[:, i], '.', label='atom {}'.format(i))
    plt.xlabel(r'Step')
    plt.ylabel(r'$x$-Position/Å')
    plt.legend(frameon=False)
    plt.show()



"""
    SMC.SolverDist.py

    Copyright (c) 2024, SAXS Team, KEK-PF    
"""
import numpy as np
import pymc as pm

def create_custom_dist(objective, lower, upper, arg_size, params):
    from Solvers.SMC.SolverLogLike import SolverLogLike

    loglike_op = SolverLogLike(objective)
    rng_ = np.random.default_rng()

    def custom_loglike(data, params):
        # data, or observed is always passed as the first input of CustomDist
        return loglike_op(params)

    def custom_random(params, rng=None, size=None):
        """
        from the original CustomDist class documentation
        https://github.com/pymc-devs/pymc/blob/main/pymc/distributions/custom.py

        see also
        Custom Distribution with pm.CustomDist
        https://discourse.pymc.io/t/custom-distribution-with-pm-customdist/11071/2

        Pickle a model containing a custom distribution
        https://discourse.pymc.io/t/pickle-a-model-containing-a-custom-distribution/9668

        """
        return rng_.uniform(lower, upper, size=arg_size)

    dummy = None
    likelihood = pm.CustomDist(
        "likelihood", params,
        observed=dummy,
        logp=custom_loglike,
        random=custom_random,
    )

    return likelihood



"""
    ModelParams.SdmSliderInfo.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.ModelParams.BaseSliderInfo import BaseSliderInfo

class SdmSliderInfo(BaseSliderInfo):
    def __init__(self, nc): 
        cmpparam_names = ["xr_h", "Rg", "uv_h"]

        cmpparam_indeces = []
        rg_base = nc + 2
        uv_base = rg_base + nc + 2
        for k in range(nc):
            cmpparam_indeces.append([k, rg_base+k, uv_base+k])

        whlparam_names = ["ma", "mb", "t0", "poresize"]
        mapping_start = nc*2 + 2
        colparam_start = nc*3 + 2 + 2 + 7 + 2
        whlparam_indeces = [mapping_start, mapping_start+1, colparam_start+2, colparam_start+3]

        BaseSliderInfo.__init__(self,
                                cmpparam_names=cmpparam_names,
                                cmpparam_indeces=cmpparam_indeces,
                                whlparam_names=whlparam_names,
                                whlparam_indeces=whlparam_indeces)


"""
    DebugPlot.py

    work-around to debug with matplotlib under tkinter

    Copyright (c) 2018-2024, Masatsuyo Takahashi, KEK-PF
"""
import matplotlib.pyplot as _plt
from matplotlib.pyplot import FuncFormatter
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
import matplotlib
import gc
# from CallStack import CallStack

def _reset():
    global dp, dp_list
    dp = None
    dp_list = []

_reset()

kill_button = False
cancel_button = True
abort_button = True
aborting = False
g_ok_text = "OK"
g_ok_only = False

def set_global_opts(**kwargs):
    global kill_button, cancel_button, abort_button, g_ok_only, g_ok_text
    ok_only = kwargs.pop('ok_only', False)
    g_ok_text = kwargs.pop('ok_text', "OK")

    if ok_only:
        g_ok_only = True
        kill_button = False
        cancel_button = False
        abort_button = False
    else:
        kill_button = kwargs.pop('kill_button', False)
        cancel_button = kwargs.pop('cancel_button', True)

g_parent = None
threaded_parent = None

def set_plot_env( parent=None, sub_parent=None ):
    global g_parent
    if g_parent is None:
        if parent is None:
            parent = sub_parent
        if parent is None:
            from molass_legacy.KekLib.TkUtils import adjusted_geometry
            parent = Tk.Tk()
            parent.geometry( adjusted_geometry( parent.geometry() ) )
            parent.withdraw()
            parent.update()
        g_parent = parent

def get_parent():
    if threaded_parent is not None:
        return threaded_parent

    set_plot_env()
    if len(dp_list) > 0:
        parent = dp_list[-1].parent
    else:
        parent = g_parent
    return parent

def get_dp():
    if len(dp_list) == 0:
        parent = get_parent()
        dp = DebugPlot(parent)
        dp_list.append(dp)
    return dp_list[-1]

def push(**kwargs):
    parent = kwargs.pop('parent', None)
    if parent is None:
        parent = get_parent()
    dp_ = DebugPlot(parent, **kwargs)
    dp_list.append(dp_)
    return dp_

def pop():
    global aborting
    dp_ = dp_list.pop()
    _plt.close(dp_.fig)
    if aborting:
        aborting = False
        assert False

class Dp:
    def __init__(self, **kwargs):
        self.kwargs = kwargs
    def __enter__(self):
        push(**self.kwargs)
    def __exit__(self, exc_type, exc_value, traceback):
        pop()

def debug_plot_close():
    # not yet successful
    if dp is not None:
        dp.destroy()

class DebugPlot( Dialog ):
    def __init__(self, parent,
                 scrollable=False,
                 window_title=None,
                 guide_message=None,
                 ok_only=None,
                 ok_text=None,
                 button_spec=None,
                 **kwargs):
        self.grab = 'local'     # used in grab_set
        self.parent = parent
        self.scrollable = scrollable
        self.mpl_canvas = None
        self.fig = None
        self.ax = None
        self.window_title = window_title
        self.ok_only = ok_only
        self.ok_text = ok_text
        self.button_spec = button_spec
        self.kwargs = kwargs
        self.guide_message = guide_message
        """
            Call to self.body must be delayed until the figure
            for FigureCanvasTkAgg become available.
            Be aware that this is not yet a Tk widget.
            E.g., you can't Dialog.destroy this instance.
        """

    def figure( self, *args, **kwargs ):
        # assert self.fig is None
        if self.fig is not None:
            # self.fig.close()
            self.fig.clf()

        # print(CallStack())
        self.fig = _plt.figure( *args, **kwargs )
        self.ax = None
        self._do_init(title=self.window_title)
        # in this case, there is no need to _do_init in self.show
        return self.fig

    def _do_init(self, title=None, visible=False, block=False):
        if title is None:
            title = "DebugPlot"
        Dialog.__init__( self, self.parent, title, visible=visible, block=block )

    def _get_fig(self):
        if self.fig is None:
            # import traceback
            # traceback.print_stack()
            self.fig = _plt.gcf()
        return self.fig

    def _get_ax(self):
        if self.ax is None:
            self.ax = self._get_fig().gca()
        return self.ax

    def show( self, block=True, pause=None ):
        """
        There are following two cases.
            Case (1)                Case (2)
            dp = DebugPlot(root)    dp = DebugPlot(root)
                                    fig = plt.figure()
                                    ax = fig.gca()
            plt.plot(...)           ax.plot()
            plt.show()              plt.show()
        """
        if self.mpl_canvas is None:
            # Case (1)
            # fig should be obtained by self._get_fig(), i.e. _plt.gcf(), in self.body
            self._do_init(title=self.window_title)
        else:
            # Case (2)
            # self._do_init() has been already done by self.figure
            pass

        self.mpl_canvas.draw()
        self.applied =False
        try:
            self._show(block=block, pause=pause)
        except:
            # _tkinter.TclError: bad window path name ".!debugplot"
            # ignore this exception for it seems harmless
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "DebugPlot._show: ")
        return self.applied

    def body(self, body_frame):     # overrides Dialog.body
        self.body_frame = body_frame
        if self.scrollable:
            from ScrolledFrame import ScrolledFrame
            button_frame = Tk.Frame(self)
            button_frame.pack(side=Tk.BOTTOM)
            sframe = ScrolledFrame(body_frame)
            bframe = sframe.interior
            sframe.pack(side=Tk.BOTTOM)
        else:
            bframe = body_frame
            button_frame = None
        cframe = Tk.Frame(bframe)
        cframe.pack()
        if self.guide_message is not None:
            label = Tk.Label(bframe, text=self.guide_message, bg="white")
            label.pack(side=Tk.TOP, fill=Tk.X)
        
        # figure should have been created before this call
        # (and before any creation of axes belonging to the figure)
        self.mpl_canvas = FigureCanvasTkAgg( self._get_fig(), cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.toolbar = NavigationToolbar( self.mpl_canvas, cframe )
        self.toolbar.update()
        if button_frame is None:
            button_frame = Tk.Frame(body_frame)
            button_frame.pack()
        self.my_buttonbox(button_frame)

    def draw(self):
        self.mpl_canvas.draw()

    def add_frame(self):
        frame = Tk.Frame(self.body_frame)
        frame.pack(side=Tk.LEFT)
        return frame

    def buttonbox(self):
        # nullify the default
        pass

    def my_buttonbox(self, frame):
        box = frame

        if self.button_spec is None:
            ok_button = True
            ok_text_ = g_ok_text
            if self.ok_text is not None:
                ok_text_ = self.ok_text

            if self.ok_only is not None and self.ok_only:
                # task: using globals is not desirable
                global cancel_button, abort_button, kill_button
                cancel_button = False
                abort_button = False
                kill_button = False
        else:
            ok_button = self.button_spec[0]
            if type(ok_button) is str:
                ok_text_ = ok_button
                ok_button = True
            elif type(ok_button) is bool:
                if ok_button:
                    ok_text_ = g_ok_text
            else:
                assert False, "unknown type of ok_button"
            if len(self.button_spec) > 1:
                cancel_button = self.button_spec[1]
            else:
                cancel_button = False
            if len(self.button_spec) > 2:
                abort_button = self.button_spec[2]
            else:
                abort_button = False
            if len(self.button_spec) > 3:
                kill_button = self.button_spec[3]
            else:
                kill_button = False

        if ok_button:
            w = Tk.Button(box, text=ok_text_, width=10, command=self.ok, default=Tk.ACTIVE)
            w.pack(side=Tk.LEFT, padx=5, pady=5)

        if cancel_button:
            w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
            w.pack(side=Tk.LEFT, padx=5, pady=5)

        if abort_button:
            w = Tk.Button(box, text="Abort", width=10, command=self.abort)
            w.pack(side=Tk.LEFT, padx=5, pady=5)

        if kill_button:
            w = Tk.Button(box, text="Kill", width=10, command=self.kill)
            w.pack(side=Tk.LEFT, padx=5, pady=5)

        extra_specs = self.kwargs.get("extra_button_specs", [])
        for name, command in extra_specs:
            width = max(10, len(name)+2)
            w = Tk.Button(box, text=name, width=width, command=command)
            w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def destroy(self):
        '''Destroy the window'''
        # print( "overriden destroy" )
        r"""
        TODO fix: there may be an instance of DebugPlot which has not been destroyed properly

          File "F:\PyTools\pytools-1_2_2-develop\lib\KekLib\OurTkinter.py", line 173, in ok
            self.withdraw()
          File "c:\program files\python36\lib\tkinter\__init__.py", line 1992, in wm_withdraw
            return self.tk.call('wm', 'withdraw', self._w)
        _tkinter.TclError: bad window path name ".!debugplot"

        SEE ALSO: 
        https://ja.osdn.net/projects/pylaf/scm/hg/pylaf/blobs/tip/src/pylafiii/mplext.py
        """
        # print( "before unbind", self.bind("<Destroy>") )
        self.unbind("<Destroy>")    # don't know whether this is the right way
        # print( "after unbind", self.bind("<Destroy>") )
        self.initial_focus = None
        Tk.Toplevel.destroy(self)

        # to avoid "RuntimeError: main thread is not in main loop"
        # see also OurTkinter.Dialog
        gc.collect()

    def apply(self):
        self.applied = True

    def abort(self):
        from tkinter import messagebox
        global aborting

        ret = messagebox.askokcancel("Abort Confirmation", "Are you sure to abort DebugPlot?")
        if ret:
            aborting = True
            self.cancel()

    def kill(self):
        import sys
        print('kill')
        self.parent.destroy()
        sys.exit()

def debug_plot_ok():
    dp = get_dp()
    return dp.applied

def figure( *args, **kwargs ):
    dp = get_dp()
    return dp.figure( *args, **kwargs )

def gcf():
    dp = get_dp()
    return dp._get_fig()

def clf():
    dp = get_dp()
    return dp._get_fig().clf()

def cla():
    dp = get_dp()
    return dp._get_ax().cla()

def gca():
    fig = figure()
    return fig.gca()

def plot( *args, **kwargs ):
    return _plt.plot( *args, **kwargs )

def show( block=True, pause=None ):
    dp = get_dp()
    ret = dp.show( block=block, pause=pause )
    return ret

def legend():
    return gcf().legend()

def tight_layout():
    return gcf().tight_layout()

def annotate( *args, **kwargs ):
    dp = get_dp()
    return dp._get_ax().annotate( *args, **kwargs )

def subplots_adjust( *args, **kwargs ):
    return _plt.subplots_adjust( *args, **kwargs )

def subplots( *args, **kwargs ):
    dp = get_dp()
    fig, axes = _plt.subplots( *args, **kwargs )
    dp.fig = fig
    dp.axes = axes
    dp._do_init()       # required to enable tk methods, e.g., dp.geometry()
    return fig, axes

def axes( *args, **kwargs ):
    return _plt.axes( *args, **kwargs )

def close():
    _plt.close()

def update():
    """
    required in animation to get button events, etc.
    """
    dp.parent.update()

def setp(*args, **kwargs):
    _plt.setp(*args, **kwargs)

class DialogWrapper:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __enter__(self):
        set_global_opts(ok_only=True, ok_text="Close")
        push(**self.kwargs)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        show()
        pop()
        set_global_opts(ok_only=False)

def get_current_dialog():
    return dp

def switch_backend(*args):
    _plt.switch_backend(*args)

"""
    see Selective.PropOptimizerUtils.py for usage
"""
def exec_in_threaded_mainloop(closure):
    global threaded_parent

    threaded_parent = Tk.Tk()
    threaded_parent.withdraw()
    threaded_parent.after(0, closure)
    threaded_parent.mainloop()

def quit_threaded_mainloop():
    global threaded_parent

    threaded_parent.quit()
    threaded_parent = None



"""
    SimpleThreedView.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt
from MatrixData import simple_plot_3d
from DataUtils import get_in_folder

def show_simple_3d_view(sd):
    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,6), subplot_kw=dict(projection="3d"))
        fig.suptitle("Simple 3D View of %s" % get_in_folder(), fontsize=20)

        ax1.set_title("UV Data", fontsize=16)
        D, _, wv, uv_curve = sd.get_uv_data_separate_ly()
        simple_plot_3d(ax1, D, x=wv)

        ax2.set_title("Xray Data", fontsize=16)
        D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
        simple_plot_3d(ax2, D, x=qv)

        fig.tight_layout()
        fig.subplots_adjust(top=0.85)
        plt.show()



"""
    Solvers.ABC.CustomPopulation.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from pyabc.population import Population, Sample, SampleFactory
import molass_legacy.KekLib.DebugPlot as plt

debug = False

class CustomSample(Sample):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_accepted_population(self) -> Population:
        """
        Returns
        -------
        population: Population
            A population of only the accepted particles.
        """
        global debug
        if debug:
            print("CustomSample: len(self.accepted_particles)=", len(self.accepted_particles))
            print("CustomSample: len(self.rejected_particles)=", len(self.rejected_particles))
            weights = [p.weight for p in self.accepted_particles]
            print("accepted_particles weights=", weights)

            with plt.Dp(button_spec=["OK", "Cancel"]):
                fig, ax = plt.subplots()
                ax.set_title("get_accepted_population")
                ax.plot(weights)
                fig.tight_layout()
                ret = plt.show()
                if not ret:
                    debug = False

        return Population(self.accepted_particles.copy())

class CustomSampleFactory(SampleFactory):
    def __call__(self, is_look_ahead: bool = False):
        """Create a new empty sample."""
        return CustomSample(
            record_rejected=self._record_rejected,
            max_nr_rejected=self._max_nr_rejected,
            is_look_ahead=is_look_ahead,
        )



"""

    Rank.SRR.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF

"""
import numpy as np
from bisect import bisect_right
from SvdDenoise import get_denoised_data
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier

class SRR:
    def __init__(self, sd):
        D, E, qv, ecurve = sd.get_xr_data_separate_ly()
        self.D = D
        self.E = E
        self.qv = qv
        self.i_smp = sd.xray_index
        self.ecurve = ecurve

    def compute_judge_info(self, figs_folder=None, analysis_name=None):
        ecurve = self.ecurve
        srr_list = []

        if figs_folder is not None:
            figs_prefix = analysis_name.replace("analysis", "fig")

        for k, peak_rec in enumerate(ecurve.get_major_peak_info()):
            print([k], peak_rec)
            start = peak_rec[0]
            stop = peak_rec[2]+1
            if start >= stop:
                # as in 20171226
                # or should be investigated for the cause
                continue

            eslice = slice(start, stop)
            D = self.D[:,eslice]
            D1_ = get_denoised_data(D, rank=1)
            D2_ = get_denoised_data(D, rank=2)
            srr1 = np.linalg.norm(D2_ - D)/np.linalg.norm(D1_ - D)

            c = self.D[self.i_smp,eslice]
            C1 = np.array([c])
            Cinv1 = np.linalg.pinv(C1)
            C2 = np.array([c, c**2])
            Cinv2 = np.linalg.pinv(C2)
            Dinv = np.linalg.pinv(D)
            E_ = self.E[:,eslice]

            sg_list = []
            rg_list = []
            qrg_list = []
            for D_, Cinv in [(D1_, Cinv1), (D2_, Cinv2)]:
                P = D_ @ Cinv
                W   = np.dot(Dinv, P)
                Pe  = np.sqrt(np.dot(E_**2, W**2))
                data = np.array([self.qv, P[:,0], Pe[:,0]]).T
                sg = SimpleGuinier(data)
                sg_list.append(sg)
                rg_list.append(max(10.0, sg.Rg))
                qrg_list.append(sg.min_qRg)

            # Q*rg = 2
            # Q = 2/rg
            rg_ = np.mean(rg_list)
            i = bisect_right(self.qv, 2.0/rg_)
            print("Rg=", rg_, "i=", i)
            aslice = slice(0, i)
            M = self.D[aslice,eslice]
            M1_ = get_denoised_data(M, rank=1)
            M2_ = get_denoised_data(M, rank=2)
            srr2 = np.linalg.norm(M2_ - M)/np.linalg.norm(M1_ - M)
            top_x = peak_rec[1]
            srr_list.append((top_x, srr1, srr2, *rg_list, *qrg_list))
            if figs_folder is not None:
                pass
                # to be implemented in SimpleGuinierDemo.py

        return srr_list



"""
    Models.Stochastic.MomentUtils.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.integrate import quad
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SecTheory.BasicModels import robust_single_pore_pdf
from molass_legacy.Peaks.ElutionModels import compute_moments_from_egh_params
from .LognormalPoreFunc import lognormal_pore_func

def compute_egh_moments(peaks):
    moments_list = []
    for params in peaks:
        tr, s, t = params[1:4]
        moments_list.append(compute_moments_from_egh_params(tr, s, t))
    return moments_list

def to_moderate_props(props):
    new_props = np.power(props, 0.5)
    return new_props/np.sum(new_props)

def moments_demo_from_rgdist(x, y, model, peaks, peak_rgs, props, monopore_params, logger=None, debug=False):
    if debug:
        from importlib import reload
        import molass_legacy.Models.Stochastic.MonoporeGuess
        reload(molass_legacy.Models.Stochastic.MonoporeGuess)
        import molass_legacy.Models.Stochastic.LognormalGuess
        reload(molass_legacy.Models.Stochastic.LognormalGuess)
    from .MonoporeGuess import guess_monopore_params_using_moments
    from .LognormalGuess import guess_lognormalpore_params_using_moments

    egh_moments_list = compute_egh_moments(peaks)
    if monopore_params is None:
        from importlib import reload
        import molass_legacy.Models.Stochastic.RoughGuess
        reload(molass_legacy.Models.Stochastic.RoughGuess)
        from .RoughGuess import guess_monopore_params_roughtly
        monopore_params = guess_monopore_params_roughtly(x, y, model, peaks, peak_rgs, props, egh_moments_list)
        if logger is not None:
            logger.info("monopore_params in moments_demo_from_rgdist: %s", monopore_params)
    print("monopore_params=", monopore_params)
    if monopore_params is None:
        return

    better_monopore_params = guess_monopore_params_using_moments(x, y, egh_moments_list, peak_rgs, props, monopore_params)
    if logger is not None:
            logger.info("better_monopore_params in moments_demo_from_rgdist: %s", better_monopore_params)
    print("better_monopore_params=", better_monopore_params)
    if True:
        lognormalpore_params = guess_lognormalpore_params_using_moments(x, y, egh_moments_list, peak_rgs, props, better_monopore_params, debug=True)
    else:
        lognormalpore_params = None
    print("lognormalpore_params=", lognormalpore_params)

    with plt.Dp(button_spec=["OK", "Cancel"]):
        fig, (ax1,ax2,ax3,ax4) = plt.subplots(ncols=4, figsize=(20,4))
        fig.suptitle("Stochastic Model Moments Study & Proof", fontsize=20)
        ax1.set_title("EGH", fontsize=16)
        ax1.plot(x, y, label='data')
        cy_list = []
        for k, params in enumerate(peaks):
            cy = model(x, params)
            cy_list.append(cy)
            ax1.plot(x, cy, ":", label='component-%d' % k)
        ty = np.sum(cy_list, axis=0)
        ax1.plot(x, ty, ":", color="red", label='model total')

        ax2.set_title("Rough Monopore", fontsize=16)

        def plot_monopore(ax, monopore_params):
            N, T, x0, me, mp, poresize = monopore_params[0:6]
            scales = monopore_params[6:]
            ax.plot(x, y, label='data')
            rho = peak_rgs/poresize
            rho[rho > 1] = 1
            cy_list = []
            for k, (r_, scale) in enumerate(zip(rho, scales)):
                ni_ = N*(1 - r_)**me
                ti_ = T*(1 - r_)**mp
                cy = scale*robust_single_pore_pdf(x - x0, ni_, ti_)
                cy_list.append(cy)
                ax.plot(x, cy, ":", label='component-%d' % k)
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red", label='model total')

        plot_monopore(ax2, monopore_params)

        ax3.set_title("Better Monopore", fontsize=16)
        plot_monopore(ax3, better_monopore_params)

        def plot_lognormalpore(ax, lognormalpore_params):
            N, T, x0, me, mp, mu, sigma = lognormalpore_params[0:7]
            scales = lognormalpore_params[7:]
            ax.plot(x, y, label='data')
            cy_list = []
            for k, (rg, scale) in enumerate(zip(peak_rgs, scales)):
                cy = lognormal_pore_func(x, scale, N, T, me, mp, mu, sigma, rg, x0)
                cy_list.append(cy)
                ax.plot(x, cy, ":", label='component-%d' % k)
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red", label='model total')

        ax4.set_title("Lognormal Pore", fontsize=16)
        if lognormalpore_params is not None:
            plot_lognormalpore(ax4, lognormalpore_params)

        fig.tight_layout()
        ret = plt.show()

    if ret:
        return better_monopore_params, lognormalpore_params
    else:
        return None


# -*- coding: utf-8 -*-
"""

    ファイル名：   GuiController.py

    処理内容：

        GUI 制御クラス

        Controller
            EntryFrame
            PilatusImageInfoTable
                PilatusImageViewer
            PreferencesDialog

        Copyright (c) 2016-2020, SAXS Team, KEK-PF

"""
import sys
import os
import re
import ExecutionWindow
import platform
from molass_legacy.KekLib.OurTkinter             import Tk, ToolTip
from molass_legacy.KekLib.TkUtils                import adjusted_geometry, split_geometry
import OurMessageBox        as     MessageBox
from GuiSettingInfo         import EntryFrame
from GuiPreferences         import PreferencesDialog
from Development            import DeveloperOptionsDialog
from molass_legacy.KekLib.TkSupplements          import BlinkingFrame, tk_set_icon_portable, SlimButton
from ImageSynthesizer       import ImageSynthesizer
from PersistentInfo         import PersistentInfo
from Preferences            import get_preference
from SynthesizerSettings    import save_settings, get_setting, set_setting
from molass_legacy.AutorgKek.AppVersion             import synthesizer_version_string
from AutoRunController      import AutoRunController

# label_bg_color = 'dim gray'
label_bg_color = 'gray25'
label_fg_color = 'white'
window_height = 600

class Controller( Tk.Toplevel ):
    def __init__( self, root, opts=None ):

        root.wm_title( synthesizer_version_string() )

        Tk.Toplevel.__init__( self, root )
        self.withdraw()

        """
            テスト制御の都合により、
            上で CommandLineOptions を import しないで、
            opts を引数で受け取る。
        """
        if opts and opts.pandastable or os.name != 'nt':
            from PilatusImageInfoTable2 import PilatusImageInfoTable
        else:
            from PilatusImageInfoTable  import PilatusImageInfoTable

        self.op_is_feasible = False
        self.op_is_manual   = True
        self.refresh_button_suggestion = None
        self.image_info_table = None
        self.geometry_resized = False

        # メイン画面を構築する。
        self.root = root
        # self.root.option_add( '*Font', get_preference( 'font' ) )
        # self.root.option_add( '*Font', ( 'TkDefaultFont', 10 ) )
        # self.root.option_add( '*Font', ( 'TkFixedFont', 9 ) )
        # self.root.option_add( '*ScrolledText.Font', ( 'DotumChe', 10 ) )
        # self.root.option_add( '*Font', ('FixedSys', 9) )
        # self.root.option_add( '*Font', ('Arial', 10) )
        # self.root.option_add( '*Font', ('DotumChe', 10) )
        # self.root.option_add( '*Font', ('Lucida Console', 9) )
        # self.root.option_add( '*Font', ('System', 10) )
        # self.root.option_add( '*Font', ('Arial Unicode MS', 10) )
        # self.root.option_add( '*Font', ('MS Sans Serif', 9) )
        # self.root.option_add( '*Font', ('Microsoft Sans Serif', 10) )
        # self.root.option_add( '*Font', ('ＭＳ Ｐゴシック', 10) )
        # self.root.option_add( '*Font', ('MS UI Gothic', 10) )
        # self.root.option_add( '*Font', ('Lucida Sans Unicode', 10) )

        tk_set_icon_portable( self, 'synthesizer' )

        menubar = Tk.Menu( self )
        self.config( menu=menubar )
        menu = Tk.Menu( menubar, tearoff=0 )
        menubar.add_cascade(label="Options", menu=menu )
        menu.add_command( label="Preferences", command=self.preferences )
        menu.add_command( label="Developer Options", command=self.developer_options )
        menu.add_command( label="Exit", command=self.quit )

        help = Tk.Menu( menubar, tearoff=0 )
#        menubar.add_cascade( label="Help", menu=help )
#        help.add_command( label="Help",     command=self.help )
#        help.add_command( label="About",    command=self.about )

        self.frame_inner = Tk.Frame( self )
        self.frame_inner.pack( side=Tk.TOP, fill=Tk.BOTH, expand=1, padx=10, pady=10 )

        setting_info_label_frame = Tk.Frame( self.frame_inner )
        setting_info_label_frame.pack( side=Tk.TOP, anchor=Tk.W )
        setting_info_label = Tk.Label( setting_info_label_frame, text= ' Setting Information Entries ', relief=Tk.FLAT, fg=label_fg_color, bg=label_bg_color )
        setting_info_label.pack( side=Tk.LEFT, anchor=Tk.W )
        self.clear_button = SlimButton( setting_info_label_frame, text='Clear', command=self.setting_clear )
        self.clear_button.pack( side=Tk.LEFT, anchor=Tk.W, padx=10 )
        clear_button_guide = Tk.Label( setting_info_label_frame, text= 'press to clear the follwing entries' )
        clear_button_guide.pack( side=Tk.LEFT, anchor=Tk.W )
        ToolTip( self.clear_button, 'Press this button to clear the setting infomation.' )

        self.entry_frame_frame = Tk.Frame( self.frame_inner )
        self.entry_frame_frame.pack()
        self.entry_frame = EntryFrame( self.entry_frame_frame, self )
        self.entry_frame.pack( side=Tk.TOP, padx=10, pady=10 )

        detector_image_info_frame = Tk.Frame( self.frame_inner )
        detector_image_info_frame.pack( side=Tk.TOP, anchor=Tk.W )

        detector_image_info_label = Tk.Label( detector_image_info_frame, text= ' Image Data Information Table ', relief=Tk.FLAT, fg=label_fg_color, bg=label_bg_color )
        detector_image_info_label.pack( side=Tk.LEFT, anchor=Tk.W )

        self.refresh_button = SlimButton( detector_image_info_frame, text='Refresh', command=self.refresh )
        self.refresh_button_guide = Tk.Label( detector_image_info_frame, text='' )

        ToolTip( self.refresh_button, 'Press this button to refresh the measured data infomation listed below.' )

        object_spec_array = []
        object_spec_array.append( [ self.refresh_button ,
                                     { "side" : Tk.LEFT, "anchor" : Tk.W, "padx" : 10 } ] )

        object_spec_array.append( [ self.refresh_button_guide,
                                     { "side" : Tk.LEFT, "anchor" : Tk.W} ] )

        self.refresh_button_suggestion = BlinkingFrame( detector_image_info_frame, object_spec_array,
                                            start_proc=self.run_button_disable,
                                            stop_proc =self.run_button_enable, debug=True )

        self.refresh_button_suggestion.pack( side=Tk.LEFT, anchor=Tk.W, padx=10 )

        self.run_button = SlimButton( detector_image_info_frame, text='Run', command=self.run, state='disabled' )
        self.run_button.pack( side=Tk.LEFT, padx=10 )
        self.run_button_guide    = Tk.Label( detector_image_info_frame, text='' )
        self.run_button_guide.pack( side=Tk.LEFT )
        ToolTip( self.run_button, 'Press this button to make synthesized images for the selected samples.' )

        table_frame = Tk.Frame( self.frame_inner )
        table_frame.pack( expand=1, fill=Tk.BOTH, padx=10, pady=10, anchor=Tk.N )

        self.synthesizer = ImageSynthesizer( table_frame )

        self.image_info_table = PilatusImageInfoTable( table_frame, self.synthesizer, self.refresh_button_suggestion )

        if self.image_info_table.num_selected_rows > 0:
            self.run_button_enable()

        # refresh_button や run_button の state を変更するかもしれないので、
        # ここで呼び出す。
        self.entry_frame.after_construction_proc()

        # getting ready to show
        self.update()
        self.deiconify()
        self.geometry( adjusted_geometry( self.geometry() ) )

        self.protocol( "WM_DELETE_WINDOW", self.quit )

    def __del__( self ):
        self.entry_frame.__del__()

    def setting_clear( self ):
        ok = MessageBox.askyesno( 
                        'Confirmation',
                        'Do you want to clear the follwing entries?'
                        )
        if not ok:
            return
        self.entry_frame.clear()

    def refresh_button_disable( self ):
        print( 'refresh_button_disable' )
        self.refresh_button.configure( state='disabled' )
        self.refresh_button_guide.configure( text='' )

    def refresh_button_enable( self ):
        print( 'refresh_button_enable' )
        self.refresh_button.configure( state='normal' )
        self.refresh_button_guide.configure( text='← be sure to press after you have changed above entries' )

    def refresh(self, autorun=False):
        if self.entry_frame.check_entries():
            need_resize = self.image_info_table.refresh( log_file_path=self.entry_frame.log_file.get(),
                                            restore_view = not self.entry_frame.has_been_changed_after_refresh(),
                                            logger=self.get_logger(), autorun=autorun )
            if need_resize:
                if not self.geometry_resized:
                    """
                    this is to fix the bug which hides all the rows that should have been shown.
                    """
                    curr_geometry = self.geometry()
                    w, h, x, y = split_geometry( curr_geometry )
                    height_increment = self.image_info_table.get_resize_height()
                    new_height = h + height_increment
                    height = min(new_height, window_height)
                    new_geometry = "%dx%d+%d+%d" % (w, height, x, y)
                    self.geometry(new_geometry)
                    self.geometry_resized = height == window_height

            self.refresh_button_suggestion.stop()
            self.entry_frame.change_reset()
        else:
            MessageBox.showerror( 'Required Entries', 'Please fill in required entries.' )

    def run_button_disable( self ):
        self.run_button.configure( state='disabled' )
        self.run_button_guide.configure( text='' )

    def run_button_enable( self ):
        self.run_button.configure( state='normal' )
        self.run_button_guide.configure( text='← press to make synthesized images' )

    def manual_buttons_disable( self ):
        self.refresh_button_disable()
        self.run_button_disable()

    def run( self ):
        self.image_info_table.do_action( 3 )

    def auto_start( self ):
        print( 'auto_start' )
        ok = MessageBox.askokcancel( 
                        'Confirmation',
                        'Do you want to start automatic control?'
                        )
        if not ok:
            return

        interval = int( self.entry_frame.watch_interval.get() )
        # 実行した watch_interval を永続化の候補とする。
        set_setting( 'watch_interval', interval )
        self.grab_set_alternative()
        self.ar_controller = AutoRunController( self, interval, self.image_info_table,
                                                log_file_path=self.entry_frame.log_file.get(),
                                                on_stop=self.grab_release_alternative )
        self.ar_controller.start()

    def grab_set_alternative( self ):
        # print( 'grab_set_alternative' )
        self.entry_frame.disable()

    def grab_release_alternative( self ):
        # print( 'grab_release_alternative' )
        self.entry_frame.enable()

    def auto_disable( self, force=False ):
        print( 'auto_disable' )
        if force or not self.op_is_manual:
            self.op_is_manual = True
            self.refresh_button_enable()
            self.run_button_enable()

    def auto_enable( self, force=False  ):
        print( 'auto_enable' )
        # TODO: check ok
        if force or self.op_is_manual:
            self.op_is_manual = False
            self.refresh()
            self.refresh_button_disable()
            self.run_button_disable()

    def quit( self ):
        if not MessageBox.askyesno("Quit", "Do you want to quit?"):
            return

        if self.entry_frame.has_been_changed_ever():
            save_policy = get_preference( 'save_policy' )
            do_save = False
            if save_policy == 'Ask':
                if MessageBox.askyesno("Entry Data Handling Question", "Do you want to save the folder/file entries?"):
                    do_save = True
            elif save_policy == 'Yes':
                do_save = True

            if do_save:
                save_settings()

        else:
            # 全く変更されていない場合は、保存しない。
            pass

        self.root.quit()    # stops mainloop
        self.destroy()

    def preferences( self ):
        if self.refresh_button_suggestion and self.refresh_button_suggestion.is_blinking():
            MessageBox.showwarning( "Not Allowed", "You can't change preferences until you refresh the image data infomation table." )
            return

        dialog = PreferencesDialog( self.root, "Usual Preferences", self.entry_frame.pilatus_counter )

        if dialog.adj_output_changed:
            self.entry_frame.__del__()
            self.entry_frame.destroy()
            self.entry_frame = EntryFrame( self.entry_frame_frame, self )
            self.entry_frame.pack( side=Tk.TOP, padx=10, pady=10 )
            self.entry_frame.after_construction_proc()

        if dialog.couter_id_changed:
            self.image_info_table.refresh(  log_file_path=self.entry_frame.log_file.get(),
                                            restore_view=True, pilatus_counter=self.entry_frame.pilatus_counter,
                                            logger=self.get_logger() )

        return

    def get_logger(self):
        return self.entry_frame.app_logger

    def developer_options( self ):
        DeveloperOptionsDialog( self.root, "Developer Options" )
        return
    """
    def help( self ):
        print( 'help' )
        return

    def about( self ):
        print( 'about' )
        return
    """



# coding: utf-8
"""
    UvData.py

    Copyright (c) 2019-2022, SAXS Team, KEK-PF
"""

import numpy as np
import logging
from bisect import bisect_right
from MatrixData import MatrixData
from SerialDataUtils import load_uv_array
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Trimming import FlowChange
from molass_legacy.UV.PlainCurve import make_secondary_e_curve_at

NORMAL_RATIO_LIMIT = 50     # < 89.5 for 20170304


class UvData(MatrixData):
    def __init__(self, folder, sd=None, array=None, vector=None):
        self.logger = logging.getLogger(__name__)
        if sd is None:
            array, vector, _, col_header = load_uv_array(folder, column_header=True)
        else:
            if sd is None:
                assert array is not None and vector is not None
            else:
                array = sd.conc_array
                vector = sd.lvector
            col_header = None
        self.vector = vector
        self.col_header = col_header
        MatrixData.__init__(self, array)
        self.set_elution_curve()
        self.check_lowest_wlen_anomaly()

    def plot(self, ax=None):
        MatrixData.plot(self, ax=ax, title="UV Absorbance", ec_color='blue')

    def set_elution_curve(self):
        pick_pos = get_setting('absorbance_picking')
        index = bisect_right(self.vector, pick_pos)
        pick_num = 1
        start = index
        stop  = index + 1
        MatrixData.set_elution_curve(self, index, slice(start, stop))

    def check_lowest_wlen_anomaly(self):
        max_y = np.max(self.data[0,:])
        ratio = max_y/self.e_curve.max_y
        if ratio > NORMAL_RATIO_LIMIT:
            # as in 20170304
            self.data[0,:] = self.data[1,:]
            self.logger.warning("UV data[0,:] have been modified due to an abnormal ratio=%.3g." % ratio)

    def set_restriction(self, xr_e_curve, flow_changes=None):
        if flow_changes is None:
            e_curve2 = make_secondary_e_curve_at(self.data, self.vector, self.e_curve, self.logger)
            fc = FlowChange(self.e_curve, e_curve2, xr_e_curve)
            flow_changes = fc.get_real_flow_changes()
        print('flow_changes=', flow_changes)
        self.i_slice = slice(0, None)
        if flow_changes[0] is None:
            flow_changes[0] = 0
        self.j_slice = slice(*flow_changes)
        return flow_changes



# coding: utf-8
"""
    Pdb.PypdbLite.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import requests

class Query:
    """
    See https://data.rcsb.org/#gql-usage-guidelines
    """
    url = "https://data.rcsb.org/graphql"

    def __init__(self, pdbid):
        self.pdbid = pdbid

    def search(self):
        query_text="""
        {
          entry(entry_id:"%s") {
            exptl {
              method
            }
          }
        }
        """
        response = requests.get(
            self.url, {"query": query_text % self.pdbid}
        )

        # print("response.status_code=", response.status_code)
        if response.status_code == 200:
            text = eval(response.text)
            # print("response.text=", text)
            return text["data"]["entry"]["exptl"]
        else:
            return None

def get_pdb_file(pdbid, filetype='pdb'):
    url = 'https://files.rcsb.org/download/'

    response = requests.get(url + pdbid + "." + filetype)
    print("response.status_code=", response.status_code)
    return response.content.decode()



"""
    BaselineInpect.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from MatrixData import simple_plot_3d
from DataUtils import get_in_folder
from importlib import reload
import Baseline.IntegralBaseline
reload(Baseline.IntegralBaseline)            
from .IntegralBaseline import IntegralBaseline

def baseline_inspect(in_folder, sd):
    print("baseline_inspect", in_folder)

    U, _, wv, uv_curve = sd.get_uv_data_separate_ly()
    M, E, qv, xr_curve = sd.get_xr_data_separate_ly()

    print(U.shape, M.shape)

    uv_stop = bisect_right(wv, 300)
    xr_stop = bisect_right(qv, 0.06)
    U_ = U[0:uv_stop,:]
    M_ = M[0:xr_stop,:]
    wv_ = wv[0:uv_stop]
    qv_ = qv[0:xr_stop]

    uv_mp = sd.absorbance.index
    xr_mp = sd.xray_index

    uv_baseline = IntegralBaseline(uv_curve.x, uv_curve.y)
    xr_baseline = IntegralBaseline(xr_curve.x, xr_curve.y)

    Ub = uv_baseline.get_baseplane(U_, 0, -1)
    Mb = xr_baseline.get_baseplane(M_, 0, -1)

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=4, figsize=(20,9), subplot_kw=dict(projection="3d"))
        fig.suptitle("Integral Baseline Inspection for %s" % get_in_folder(in_folder),fontsize=20)
        ax1 = axes[0,0]
        ax2 = axes[0,1]
        ax3 = axes[0,2]
        ax4 = axes[0,3]
        ax5 = axes[1,0]
        ax6 = axes[1,1]
        ax7 = axes[1,2]
        ax8 = axes[1,3]

        ax1.set_title("UV whole",fontsize=16)
        ax2.set_title("UV head",fontsize=16)
        ax3.set_title("UV head baseline",fontsize=16)
        ax4.set_title("UV head corrected",fontsize=16)
        ax5.set_title("XR whole",fontsize=16)
        ax6.set_title("XR head",fontsize=16)
        ax7.set_title("XR head baseline",fontsize=16)
        ax8.set_title("XR head corrected",fontsize=16)
        simple_plot_3d(ax1, U, x=wv)
        simple_plot_3d(ax5, M, x=qv)
        simple_plot_3d(ax2, U_, x=wv_)
        simple_plot_3d(ax6, M_, x=qv_)

        uv_zlim = ax2.get_zlim()
        for ax in ax3, ax4:
            ax.set_zlim(uv_zlim)

        xr_zlim = ax6.get_zlim()
        for ax in ax7, ax8:
            ax.set_zlim(xr_zlim)

        simple_plot_3d(ax3, Ub, x=wv_)
        simple_plot_3d(ax4, U_ - Ub, x=wv_)
        simple_plot_3d(ax7, Mb, x=qv_)
        simple_plot_3d(ax8, M_ - Mb, x=qv_)

        uv_y = uv_curve.x
        uv_z = uv_curve.y
        uv_b = uv_baseline.yb
        uv_x = np.ones(len(uv_y))*wv[uv_mp]
        for ax in ax1, ax2:
            ax.plot(uv_x, uv_y, uv_z, color="blue")
            ax.plot(uv_x, uv_y, uv_b, color="red", alpha=0.5)

        xr_y = xr_curve.x
        xr_z = xr_curve.y
        xr_b = xr_baseline.yb
        xr_x = np.ones(len(xr_y))*qv[xr_mp]
        for ax in ax5, ax6:
            ax.plot(xr_x, xr_y, xr_z, color="green")
            ax.plot(xr_x, xr_y, xr_b, color="red", alpha=0.5)

        fig.tight_layout()
        plt.show()



# coding: utf-8
"""
    ProbDensity.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from sklearn.mixture import GaussianMixture
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.SciPyCookbook import smooth
from .ProbDensityUtils import generate_sample_data

N_CLUSTERS_FOR_PRE_RECOG = 10
MAX_ITER_FOR_PRE_RECOG = 50     # 20 is too small, i.e., unstable

class ProbDensity:
    def __init__(self, y, random_state=None):
        self.y = y
        self.N = N = len(y)
        self.x = x = np.arange(N)
        self.sy = y_ = smooth(y)
        self.py = py = y_.copy()
        py[y_ < 0] = 0
        self.M0 = np.sum(self.py)
        self.ny = self.py/self.M0
        self.pdf = UnivariateSpline(x, self.ny, s=0, ext=3)
        self.spline = UnivariateSpline(x, self.py, s=0, ext=3)
        self.data = generate_sample_data(py, 2)
        self.M1 = np.sum(x*py)/self.M0
        self.M2 = np.sum(self.py*(x - self.M1)**2)/self.M0
        self.s = np.sqrt(self.M2)
        k = int(N*0.99)
        pp = np.argpartition(py, k)
        self.mode = np.average(pp[k:])
        print('m, s=', self.M1, self.s)

    def proof_plot(self, ax=None, show_np_params=False):

        if ax is None:
            plt.push()
            fig  = plt.figure()
            ax_ = fig.gca()
        else:
            ax_ = ax

        ax_.plot(self.y, alpha=0.5)
        ax_.plot(self.sy)

        ymin, ymax =ax_.get_ylim()
        ax_.set_ylim(ymin, ymax)

        m = self.M1
        ax_.plot([m, m], [ymin, ymax], color='green')
        s = self.s
        for x in [m-s, m+s]:
            ax_.plot([x, x], [ymin, ymax], color='orange')

        if show_np_params:
            m_ = np.mean(self.data)
            ax_.plot([m_, m_], [ymin, ymax], ':', color='green')

            s_ = np.std(self.data)
            for x in [m-s_, m+s_]:
                ax_.plot([x, x], [ymin, ymax], ':', color='orange')

        if ax is None:
            fig.tight_layout()
            plt.show()
            plt.pop()

    def get_approx_peak_range(self, nsigma=10):
        m = self.M1
        s = self.s
        return max(0, int(m - s*nsigma)), min(self.N, int(m + s*nsigma))



# coding: utf-8
"""
    SvdTutorial.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
import matplotlib.pyplot    as plt
from mpl_toolkits.mplot3d   import Axes3D
from matplotlib.patches     import Circle, Ellipse
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from CanvasFrame            import CanvasFrame
from OurMatplotlib3D        import pathpatch_2d_to_3d, pathpatch_translate

DEBUG   = False

class SvdTutorial( Dialog ):
    def __init__( self, parent, array, indeces ):
        self.parent = parent
        self.array  = array
        print( self.array.shape )
        self.Qindex = indeces[0]
        self.Eindex = indeces[1]
        self.M  = array[self.Eindex,:,1][:,self.Qindex].T
        print( self.M.shape )
        U, s, VT = np.linalg.svd( self.M )
        print( U )
        print( s )
        print( VT )
        self.U  = U
        self.s  = s
        self.VT = VT

    def show( self ):
        Dialog.__init__( self, self.parent, "SVD Tutorial Demo" )

    def body( self, body_frame ):   # overrides parent class method

        base_frame = Tk.Frame( body_frame );
        base_frame.pack( expand=1, fill=Tk.BOTH )

        figsize = ( 12, 11 ) if is_low_resolution() else ( 12, 11 )
        self.canvas_frame = canvas_frame = CanvasFrame( base_frame, figsize=figsize )
        canvas_frame.pack( side=Tk.LEFT )

        panel_frame = Tk.Frame( base_frame )
        panel_frame.pack( side=Tk.LEFT, anchor=Tk.N )

        fig = canvas_frame.fig

        self.axes = []
        for i in range(2):
            axes_row = []
            for j in range(2):
                projection = '3d' if j==1 else None
                ax = fig.add_subplot( 220 + i*2 + (j+1), projection=projection )
                axes_row.append( ax )
            self.axes.append( axes_row )

        fig.tight_layout()

        self.elution_canvas = CanvasFrame( panel_frame, figsize=( 6, 8.25 ) )
        self.elution_canvas.pack()

        fig = self.elution_canvas.fig

        self.p_ax0 = fig.add_subplot( 311 )
        self.p_ax1 = fig.add_subplot( 312 )
        self.p_ax2 = fig.add_subplot( 313 )

        fig.tight_layout()

        self.draw_panel()
        self.draw_svd()

    def draw_panel( self ):
        self.p_ax0.set_title( "210-d Elution Vectors" )
        self.p_ax1.set_title( "210-d Elution Vector at Q[%d]" % self.Qindex[1] )
        self.p_ax2.set_title( "1176-d Scattering Vectors" )

        e_size = self.array.shape[0]

        y66 = np.zeros( e_size )
        y66[66] = 1

        y141 = np.zeros( e_size )
        y141[141] = 1

        ym = np.zeros( e_size )
        ym[66] = 0.5
        ym[141] = 0.5

        for y in [ y66, y141, ym ]:
            self.p_ax0.plot( y )

        for ax in [ self.p_ax0, self.p_ax1 ]:
            ax.plot( self.array[ :, self.Qindex[1], 1 ], color='orange' )

        ax = self.p_ax1
        ymin, ymax = ax.get_ylim()
        ax.set_ylim( ymin, ymax )
        for j in self.Eindex:
            ax.plot( [ j, j ], [ ymin, ymax ], ':', color='red' )

        quarter = self.array.shape[1]//4
        my66    = self.array[ self.Eindex[0], 0:quarter, 1 ]
        my141   = self.array[ self.Eindex[1], 0:quarter, 1 ]
        self.p_ax2.plot( my66 )
        self.p_ax2.plot( my141 )
        self.p_ax2.plot( (my66 + my141)/2 )

        ymin, ymax = self.p_ax2.get_ylim()
        self.p_ax2.set_ylim( ymin, ymax )
        for k in self.Qindex:
            self.p_ax2.plot( [ k, k ], [ ymin, ymax ], ':', color='red' )

    def draw_svd( self ):
        axE = self.axes[0][0]
        axS = self.axes[0][1]
        axF = self.axes[1][0]
        axG = self.axes[1][1]

        # Elution Curve Space
        axE.set_title( "Elution Curve Space", y=0.95 )
        axE.set_xlabel( "E%d" % self.Eindex[0])
        axE.set_ylabel( "E%d" % self.Eindex[1] )

        Epoints = []
        labels = []

        label = "E%d" % self.Eindex[0]
        axE.plot( 1, 0, 'o', label=label )
        Epoints.append( [1,0] )
        labels.append( label )

        label = "E%d" % self.Eindex[1]
        axE.plot( 0, 1, 'o', label=label )
        Epoints.append( [0,1] )
        labels.append( label )

        label = "Emiddle"
        axE.plot( 0.5, 0.5, 'o', label=label )
        Epoints.append( [0.5,0.5] )
        labels.append( label )

        scale = 1

        for i, q in enumerate(self.Qindex):
            x   = self.M[i,0]*scale
            y   = self.M[i,1]*scale
            label = "Q%d" % q 
            axE.plot( x, y, 'o', label=label)
            Epoints.append( [x,y] )
            labels.append( label )

        circle = Circle((0, 0), 1.0, alpha=0.1)
        axE.add_patch( circle )

        axE.plot( [1,0], [0,1], ':', color='black' )
        axE.plot( 0, 0, 'o', color='red', label='zero' )

        axE.legend()

        # Scattering Curve Space
        axS.set_title( "Scattering Curve Space", y=1.03 )
        axS.set_xlabel( "Q%d" % self.Qindex[0])
        axS.set_ylabel( "Q%d" % self.Qindex[1] )
        axS.set_zlabel( "Q%d" % self.Qindex[2] )

        Spoints = []
        for k, p in enumerate(Epoints):
            S   = np.dot( self.M, np.array(p).reshape( (2,1) )*scale ).reshape( (3,) )
            axS.plot( [ S[0] ], [ S[1] ], [ S[2] ], 'o', label=labels[k] )
            Spoints.append( [ S[0], S[1], S[2] ] )

        Spoints_ = np.array( Spoints )
        axS.plot( Spoints_[0:2,0], Spoints_[0:2,1], Spoints_[0:2,2], ':', color='black' )
        axS.plot( [0], [0], [0], 'o', color='red', label='zero' )

        p = Ellipse((0,0), self.s[0]*2, self.s[1]*2, alpha=0.1 )    # Add an ellipse in the xy plane
        axS.add_patch(p)

        normal = np.dot( self.U, [0, 0, -11] ).reshape( (3,) )
        print( 'normal=', normal )
        pathpatch_2d_to_3d(p, z = 0, normal=normal)

        axS.legend()

        # Roteted Elution Curve Space
        axF.set_title( "Roteted Elution Curve Space", y=0.95 )
        Fpoints = []
        for k, p in enumerate(Epoints):
            F   = np.dot( self.VT, np.array(p).reshape( (2,1) )*scale )
            axF.plot( F[0,0], F[1,0], 'o', label=labels[k] )
            Fpoints.append( [ F[0,0], F[1,0] ] )

        Fpoints_ = np.array( Fpoints )

        circle = Circle((0, 0), 1.0, alpha=0.1)
        axF.add_patch( circle )
        axF.plot( Fpoints_[0:2,0], Fpoints_[0:2,1], ':', color='black' )
        axF.plot( 0, 0, 'o', color='red', label='zero' )

        axF.legend()

        # Scaled Elution Curve Space embedded into Scattering Curve Space
        axG.set_title( "Scaled R-EC Space embedded into SC Space", y=1.03 )

        p = Ellipse((0,0), self.s[0]*2, self.s[1]*2, alpha=0.1 )    # Add an ellipse in the xy plane
        axG.add_patch(p)
        pathpatch_2d_to_3d(p, z = 0, normal = 'z')

        Gpoints = []
        for k, p in enumerate(Fpoints):
            p_  = self.s * p
            axG.plot( [ p_[0] ], [ p_[1] ], [0], 'o', label=labels[k] )
            Gpoints.append( p_ )

        Gpoints_ = np.array( Gpoints )

        axG.plot( Gpoints_[0:2,0], Gpoints_[0:2,1], [0], ':', color='black' )
        axG.plot( [0], [0], [0], 'o', color='red', label='zero' )

        axG.legend()

        if False:
            # check whether sel.U transforms Gpoints[1] to the right position
            p   = Gpoints[1]
            Sd  = np.dot( self.U, [p[0], p[1], 0] ).reshape( (3,) )
            axS.plot( [Sd[0]], [Sd[1]], [Sd[2]], 'o', color='yellow', label='debug' )

        xyz_lim = np.max( self.s )

        for ax in [ axS, axG ]:
            ax.set_xlim( -xyz_lim, xyz_lim )
            ax.set_ylim( -xyz_lim, xyz_lim )
            ax.set_zlim( -xyz_lim, xyz_lim )

        self.canvas_frame.show()
        


"""

    PopupMenuUtils.py

    Copyright (c) 2024, Masatsuyo Takahashi, KEK-PF

"""
def post_popup_menu(popup_menu, anchor_widget, event, mpl_event=False, debug=False):
    from molass_legacy.KekLib.TkUtils import split_geometry

    cx = anchor_widget.winfo_rootx()
    cy = anchor_widget.winfo_rooty()
    w, h, x, y = split_geometry(anchor_widget.winfo_geometry())

    if debug:
        print("cx, cy=", cx, cy)
        print("w, h, x, y=", w, h, x, y)
        print("event.x, event.y=", event.x, event.y)

    if mpl_event:
        popup_menu.post(cx + int(event.x), cy + h - int(event.y))
    else:
        # currently adjusted only for VpaButtonFrame
        # why + 100, +20 ?
        popup_menu.post(cx + x + int(event.x) + 100, cy + int(event.y) + 20)

if __name__ == '__main__':
    pass



# coding: utf-8
"""
    Rdf.py

    Copyright (c) 2020-2022, SAXS Team, KEK-PF
"""
import os
import numpy as np
from scipy.integrate import quad
from scipy.interpolate import UnivariateSpline
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.NumpyUtils import np_loadtxt_robust
from molass_legacy.AutorgKekAdapter import AutorgKekAdapter
from molass_legacy.ATSAS.DatGnom import DatgnomExecutor

class Rdf:
    def __init__(self, q, I):
        self.interp = UnivariateSpline(q, I, s=0, ext=1)
        self.q = q
        self.vec_int = np.vectorize(self.expint)

    def integrand(self, t, r):
        return self.interp(t)*t*r*np.sin(t*r)

    def expint(self, r):
        return 4*np.pi*quad(self.integrand, 0, self.q[-1], args=(r))[0]

    def compute(self, r):
        return self.vec_int(r)

class AtsasDatGnomDdf:
    def __init__(self, path):
        self.datgnom = DatgnomExecutor()
        self.path = path
        self.array, _ = np_loadtxt_robust(path)
        self.q = self.array[:,0]
        self.y = self.array[:,1]
        autorg_kek = AutorgKekAdapter( self.array )
        self.guinier_result = autorg_kek.run()

    def guess_best_dmax(self):
        temp_folder = get_setting('analysis_folder') + '/.temp'
        if not os.path.exists(temp_folder):
            from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
            mkdirs_with_retry(temp_folder)
        self.out_file = temp_folder + '/datgnom_temp.out'
        rg = self.guinier_result.Rg
        self.result = self.datgnom.execute(self.path, rg, self.out_file)
        ret = self.get_pr_curve()
        r = ret[:,5]
        dmax = r[np.isfinite(r)][-1]
        return dmax

    def get_pr_curve(self, debug=False):
        from molass_legacy.ATSAS.DatGnom import datgnom_read_data

        def positive_value( v ):
            v_ = float(v)
            return v_ if v_ > 0 else np.nan

        exper_fit, real_space = datgnom_read_data(self.out_file, null_value=np.nan, null_func=positive_value)

        num_rows = max(len(exper_fit), len(real_space))
        rows = []
        for i in range(num_rows):
            if i < len(exper_fit):
                row = exper_fit[i]
            else:
                row = [np.nan] * 5
            if i < len(real_space):
                row += real_space[i]
            else:
                row += [np.nan] * 3
            rows.append( row )

        ret = np.array(rows)

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            fig, axes = plt.subplots(nrows=1,  ncols=2, figsize=(14, 6))
            ax1, ax2 = axes

            s = ret[:,0]
            jexp = ret[:,1]
            iexp = ret[:,4]
            r = ret[:,5]
            pr = ret[:,6]

            ax1.plot(s, np.log10(jexp))
            ax1.plot(s, np.log10(iexp))
            ax2.plot(r, pr)

            fig.tight_layout()
            plt.show()

        return ret



# coding: utf-8
"""
    BaseSurfaceDemo.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt
from MatrixData import simple_plot_3d
from MeasuredData import MeasuredData
from DataUtils import get_in_folder

def figure_for_doc(in_folder):
    print(in_folder)
    md = MeasuredData(in_folder)
    # md.plot()

    xd = md.xr

    plt.push()
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.set_title("Non-Corrected Data", fontsize=20)
    simple_plot_3d(ax, xd.data, x=xd.vector, alpha=0.3)
    fig.tight_layout()
    plt.show()
    plt.pop()



# coding: utf-8
"""
    EmgMixtureUtils.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from sklearn.mixture import GaussianMixture
from lmfit import minimize, Parameters
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
import molass_legacy.KekLib.DebugPlot as dplt
from .ProbDataUtils import generate_samle_datasets, plot_hist_data
from .EmgMixture import emg_pdf

def get_gmm(size, data, k, max_iter=100, anim_data=False):
    from .EmgMixture import EmgMixture
    gmm = EmgMixture(k, max_iter=max_iter, anim_data=anim_data)
    gmm.fit(X=np.expand_dims(data,1), bins=size)
    return gmm

def set_consistent_base(ax1, axt):
    ymin1, ymax1 = ax1.get_ylim()
    ymint, ymaxt = axt.get_ylim()
    y_ = ymin1/ymax1*ymaxt
    axt.set_ylim(y_, ymaxt)

def gaussian(x, mu, sigma, scale):
    return scale * np.exp( - (x - mu)**2 / (2 * sigma**2) )

def get_components(x, y, gmm):
    gy_list = []
    ty = np.zeros(len(x))
    for k in range(gmm.K):
        w = gmm.pi[k]
        m = gmm.mu[k]
        s = gmm.sigma[k]
        t = gmm.tau[k]
        print([k], w, m, s, t)
        gy = emg_pdf(x, m, s, t, w)
        gy_list.append([gy, m])
        ty += gy

    sorted_gy_list = sorted(gy_list, key=lambda x:x[1])

    if False:
        dplt.push()
        fig = dplt.figure()
        ax = fig.gca()
        ax.plot(ty)
        for k, rec in enumerate(sorted_gy_list):
            gy = rec[0]
            ax.plot(gy, ':', label=str(k))

        ax.legend()
        fig.tight_layout()
        dplt.show
        dplt.pop()

    def obj_func(params):
        S   = params['S']
        return ty*S - y

    params = Parameters()
    S_init = np.max(y)/np.max(ty)
    params.add('S', value=S_init, min=0, max=S_init*100 )
    result = minimize(obj_func, params, args=())

    scale = result.params['S'].value
    return [ scale*rec[0] for rec in sorted_gy_list ], ty*scale

def plot_components(ax, x, cgys):
    ret_lines = []
    for k, gy in enumerate(cgys):
        line, = ax.plot(x, gy, ':', label='component-%d' % k)
        ret_lines.append(line)
    return ret_lines

def plot_gmm_elution(y1, y2, data1, data2, k, max_iter=100, axes=None, plot_class=plt, anim_data=False):
    from DataUtils import get_in_folder

    gmm1 = get_gmm(len(y1), data1, k, max_iter=max_iter, anim_data=anim_data)
    gmm2 = get_gmm(len(y2), data2, k, max_iter=max_iter, anim_data=anim_data)

    if axes is None:
        fig, axes = plot_class.subplots(nrows=1, ncols=2, figsize=(14,7))
        fig.suptitle("Decomposition of %s using EMG Mixture Model" % get_in_folder(), fontsize=20)

    ax1, ax2 = axes

    ax1.set_title("UV Elution Decomposition", fontsize=16)
    ax2.set_title("X-ray Elution Decomposition", fontsize=16)
    ax1.plot(y1, label='input data')
    ax2.plot(y2, label='input data')

    x1 = np.arange(len(y1))
    cgys1, gy1 = get_components(x1, y1, gmm1)
    ax1.plot(gy1, label='emgmm-fit')
    plot_components(ax1, x1, cgys1)

    x2 = np.arange(len(y2))
    cgys2, gy2 = get_components(x2, y2, gmm2)
    ax2.plot(gy2, label='eghmm-fit')
    plot_components(ax2, x2, cgys2)

    ax1.legend()
    ax2.legend()

    if axes is None:
        fig.tight_layout()
        fig.subplots_adjust(top=0.88)

    return gmm1, gmm2

def get_anim_components(x, y, gmm, n):
    gy_list = []
    ty = np.zeros(len(x))
    for k in range(gmm.K):
        w = gmm.pi_array[n,k]
        m = gmm.mu_array[n,k]
        s = gmm.sigma_array[n,k]
        t = gmm.tau_array[n,k]
        # print([n,k], w, m, s, t)
        gy = emg_pdf(x, m, s, t, w)
        gy_list.append([gy, m])
        ty += gy

    sorted_gy_list = sorted(gy_list, key=lambda x:x[1])

    if True:
        dplt.push()
        fig = dplt.figure()
        ax = fig.gca()
        ax.plot(ty)
        for k, rec in enumerate(sorted_gy_list):
            gy = rec[0]
            ax.plot(gy, ':', label=str(k))

        ax.legend()
        fig.tight_layout()
        dplt.show
        dplt.pop()

    def obj_func(params):
        S   = params['S']
        return ty*S - y

    params = Parameters()
    S_init = np.max(y)/np.max(ty)
    params.add('S', value=S_init, min=0, max=S_init*100 )
    result = minimize(obj_func, params, args=())

    scale = result.params['S'].value
    return [ scale*rec[0] for rec in sorted_gy_list ], ty*scale

def animate_gmm_elution(y1, y2, gmm1, gmm2, plot_class=plt):
    from DataUtils import get_in_folder

    # fig, axes = plot_class.subplots(nrows=1, ncols=2, figsize=(14,7))     # does not work for animation
    # ax1, ax2 = axes

    fig = plot_class.figure(figsize=(14,7))
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)

    fig.suptitle("Decomposition of %s using EMG Mixture Model Animation" % get_in_folder(), fontsize=20)
    ax1.set_title("UV Elution Decomposition", fontsize=16)
    ax2.set_title("X-ray Elution Decomposition", fontsize=16)
    ax1.plot(y1, label='input data')
    ax2.plot(y2, label='input data')

    x1 = np.arange(len(y1))
    cgys1, gy1 = get_anim_components(x1, y1, gmm1, 0)
    line1, = ax1.plot(gy1, label='emgmm-fit')
    lines1 = plot_components(ax1, x1, cgys1)

    x2 = np.arange(len(y2))
    cgys2, gy2 = get_anim_components(x2, y2, gmm2, 0)
    line2, = ax2.plot(gy2, label='emgmm-fit')
    lines2 = plot_components(ax2, x2, cgys2)

    def add_number_text(ax):
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        tx = (xmin + xmax)/2
        ty = (ymin + ymax)/2
        text = ax.text(tx, ty, "_", ha='center', va='center', alpha=0.05, fontsize=240)
        return text

    text1 = add_number_text(ax1)
    text2 = add_number_text(ax2)

    ax1.legend()
    ax2.legend()
    fig.tight_layout()
    fig.subplots_adjust(top=0.88)

    anim_artists = [line1] + lines1 + [line2] + lines2 + [text1, text2]

    def animate(i):
        # print('animate', i)

        cgys1, gy1 = get_anim_components(x1, y1, gmm1, i+1)
        line1.set_data(x1, gy1)
        for line, gy in zip(lines1, cgys1):
            line.set_data(x1, gy)

        cgys2, gy2 = get_anim_components(x2, y2, gmm2, i+1)
        line2.set_data(x2, gy2)
        for line, gy in zip(lines2, cgys2):
            line.set_data(x2, gy)

        text_num = str(i+1)
        text1.set_text(text_num)
        text2.set_text(text_num)

        dplt.update()   # need a chance to get events

        return anim_artists

    def reset():
        print('reset')
        return animate(-1)

    num_frames = gmm1.num_iter
    print('num_frames=', num_frames)
    anim = animation.FuncAnimation(fig, animate, frames=num_frames, blit=True, init_func=reset, interval=100)

    return anim

def spike_demo(in_folder, **kwargs):
    n_components = kwargs.pop('n_components', 2)
    max_iter = kwargs.pop('max_iter', 100)
    anim_data = kwargs.pop('anim_data', False)

    print("max_iter=", max_iter)

    y1, y2, data1, data2 = generate_samle_datasets(in_folder, kwargs)

    dplt.push()
    plot_hist_data(y1, y2, data1, data2, plot_class=dplt)
    dplt.show()
    dplt.pop()

    dplt.push()
    gmm1, gmm2 = plot_gmm_elution(y1, y2, data1, data2, n_components, max_iter=max_iter, plot_class=dplt, anim_data=anim_data)
    dplt.show()
    dplt.pop()

    if anim_data:
        dplt.push()
        animate_gmm_elution(y1, y2, gmm1, gmm2, plot_class=dplt)
        dplt.show()
        dplt.pop()

class SpikeDialog(Dialog):
    def __init__(self, parent, in_folder, **kwargs):
        self.continue_ = True
        Dialog.__init__(self, parent, title="SpikeDialog", visible=False)
        self.generate_samle(in_folder, **kwargs)

    def body(self, bframe):
        self.cframe = Tk.Frame(bframe)
        self.cframe.pack()

    def buttonbox( self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Cancel", width=10, command=self._cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        w = Tk.Button(box, text="Next", width=10, command=self.draw_next)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        w = Tk.Button(box, text="Iterate", width=10, command=self.iterate)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def _cancel(self):
        self.continue_ = False
        self.cancel()

    def generate_samle(self, in_folder, **kwargs):
        y1, y2, data1, data2 = generate_samle_datasets(in_folder, kwargs)

        if False:
            np.savetxt('y1.dat', y1)
            np.savetxt('y2.dat', y2)

            np.savetxt('data1.dat', data1[:100])
            np.savetxt('data2.dat', data2[:100])

        cframe = self.cframe

        self.fig, self.axes = plt.subplots(nrows=2, ncols=2, figsize=(12,8))
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, cframe)
        self.toolbar.update()

        plot_hist_data(y1, y2, data1, data2, axes=self.axes[0,:])
        n_components = kwargs.pop('n_components', 2)
        self.mm_args = [y1, y2, data1, data2, n_components]
        self.kwargs = kwargs
        self.draw_next()
        self.fig.tight_layout()
        self.mpl_canvas.draw()

    def show(self):
        self._show()

    def draw_next(self):
        axes = self.axes[1,:]
        for ax in axes:
            ax.cla()
        plot_gmm_elution(*self.mm_args, axes=axes, **self.kwargs)
        self.mpl_canvas.draw()

    def iterate(self):
        while self.continue_:
            self.draw_next()
            time.sleep(1)
            self.update()

def spike_dialog(parent, in_folder, **kwargs):
    dialog = SpikeDialog(parent, in_folder, **kwargs)
    dialog.show()



"""
    DenssManager.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF
"""
from time import time, sleep
from .DenssUtils import get_outfolder
from molass_legacy._MOLASS.SerialSettings import get_setting

class JobInfo:
    def __init__(self, code, **kwargs):
        self.code = code
        self.analysis_name = None
        self.__dict__.update(kwargs)
        if self.analysis_name is None:
            self.analysis_name = get_setting('analysis_name')

    def __repr__(self):
        return 'JobInfo<%s>' % self.code

denss_manager = None

def activate_manager(log_folder, use_gpu):
    global denss_manager
    if denss_manager is None:
        denss_manager = DenssManager(log_folder, use_gpu)
    return denss_manager.num_workers

def terminate_manager_impl():
    if denss_manager is not None:
        denss_manager.terminate()

def print_table():
    print(denss_manager.table)

def get_table():
    return denss_manager.table

def get_list():
    ret = []
    if denss_manager is not None:
        for k, item in sorted(denss_manager.table.items()):
            ret.append((k, item))
    return ret

def submit(job_info):
    job_id = denss_manager.accept_job(job_info)
    return job_id

class DenssManager:
    def __init__(self, log_folder, use_gpu):
        from multiprocessing import Pool, Manager
        from molass_legacy.KekLib.BasicUtils import rename_existing_file
        from molass_legacy.KekLib.ChangeableLogger import Logger
        log_path = log_folder+'/denss_manager.log'
        rename_existing_file(log_path, ext=".log")
        self.logger = Logger(log_path)
        self.use_gpu = use_gpu
        self.job_counter = 0
        self.manager = Manager()
        self.table = self.manager.dict({})
        self.num_workers = self.decide_num_workers()
        self.pool = Pool(processes=self.num_workers)
        self.logger.info('started with number of workers %d', self.num_workers)


    def decide_num_workers(self):
        import os
        cores = os.cpu_count()//2
        return max(1, cores//2)

    def accept_job(self, job_info):
        from .DenssDriver import run_denss
        job_id = self.job_counter
        self.job_counter += 1
        submitted = time()
        file_info = [job_info.analysis_name, job_info.infile_name]
        self.table[job_id] = [file_info, None, -2, -1, None, None, None, submitted, None, None, None, 0]
        def success_callback_(result):
            self.success_callback(job_id, result)

        def error_callback_(exc_obj):
            self.error_callback(job_id, exc_obj)

        out_folder = get_outfolder(job_id)
        self.pool.apply_async(run_denss, (job_id, out_folder, job_info, self.table, self.use_gpu), {}, success_callback_, error_callback_)
        self.logger.info('accepted %s as %d', str(job_info), job_id)
        return job_id

    def success_callback(self, job_id, result):
        self.logger.info('%03d %s', job_id, str(result))

    def error_callback(self, job_id, exc_obj):
        self.logger.error('%03d error %s', job_id, str(exc_obj))

    def terminate(self):
        self.pool.close()
        self.pool.join()



# coding: utf-8
"""

    ElutionalSmootherUtils.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""

class LightIntensityData:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)



# coding: utf-8
"""

    ResultTable.py

    Copyright (c) 2016-2017, Masatsuyo Takahashi, KEK-PF

"""

import os
from molass_legacy.KekLib.OurTkinter             import Tk, ToolTip
import OurMessageBox        as     MessageBox
from TkMiniTable            import TkMiniTable
from molass_legacy._MOLASS.SerialSettings               import get_setting

NUM_RESULT_COLUMNS  = 28
QUALITY_SIGNAL      =  7
MONO_RG             = 12
IPI_FLAG            = 19
BICO_FLAG           = 20

class ResultTable:
    def __init__( self, parent, rec_array ):
        self.parent = parent
        self.table  = None

        self.refresh( rec_array )

    def refresh( self, rec_array ):

        if self.table is not None:
            self.table.destroy()

        columns = list( map( lambda n: '%02d' % n, range( NUM_RESULT_COLUMNS ) ) )
        self.table = table = TkMiniTable( self.parent, columns=columns, default_colwidth=8 )
        table.pack( fill=Tk.BOTH, anchor=Tk.NW  )

        table.heading( '00', 'Tool ID.' )
        table.column ( '00', width=20 )
        table.heading( '01', 'Rg' )
        table.heading( '02', 'Rg StDev' )
        table.heading( '03', 'I(0)' )
        table.heading( '04', 'I(0) StDev' )
        table.heading( '05', 'First Point' )
        table.heading( '06', 'Last Point' )
        table.heading( '07', 'Quality' )
        table.heading( '08', 'Aggregated' )
        table.heading( '09', 'min q*Rg' )
        table.heading( '10', 'max q*Rg' )
        table.heading( '11', 'Mono G' )
        table.column ( range( 11, 14 ), bg = 'ivory')
        table.heading( '12', 'Mono Rg' )
        table.heading( '13', 'Mono d' )
        table.heading( '14', 'basic qual' )
        table.column ( range( 14, 19 ), bg = 'honeydew')
        table.heading( '15', 'positive s' )
        table.heading( '16', 'consistenc' )
        table.heading( '17', 'stdev scor' )
        table.heading( '18', 'qRg score' )
        table.heading( '19', 'IpI flag' )
        table.column ( range( 19, 28 ), bg = 'seashell')
        table.heading( '20', 'BiCo flag' )
        table.heading( '21', 'b/m ratio' )
        table.heading( '22', 'BiCo G1' )
        table.heading( '23', 'BiCo G2' )
        table.heading( '24', 'BiCo Rg1' )
        table.heading( '25', 'BiCo Rg2' )
        table.heading( '26', 'BiCo d1' )
        table.heading( '27', 'BiCo d2' )

        rec_a = rec_array[0]
        try:
            atsas_signal = 'white' if rec_a is None else rec_a.get_atsas_signal()
        except:
            # ErrorResult
            atsas_signal = 'white'
        rec_k = rec_array[2]
        try:
            kekpf_signal = 'white' if rec_k is None else rec_k.get_kekpf_signal()
        except:
            # ErrorResult
            kekpf_signal = 'white'

        signals     = [ atsas_signal, None, kekpf_signal ]
        tool_list   = [ 'ATSAS original', 'ATSAS re-evaluated', 'KEK' ]

        for row, rec in enumerate( rec_array ):
            val_array = []
            val_array.append( tool_list[row] )

            if rec is None:
                val_array += [ '' ] * ( NUM_RESULT_COLUMNS - 1 )
            else:
                for i in list( range(NUM_RESULT_COLUMNS - 1) ):
                    try:
                        if rec[i] is None:
                            strval = 'NA'
                        else:
                            strval = '%.4g' % rec[i]
                    except:
                        # ErrorResult
                        strval = 'NA'

                    val_array.append( strval )

            # print( val_array )
            table.insert_row( val_array )

            color = signals[ row ]
            if color is not None:
                fg = 'black' if color == 'yellow' else 'white'
                table.cells_array[ row+1, QUALITY_SIGNAL ].widget.config( bg=color, fg=fg )

        for col, color in [ [ MONO_RG, 'light blue' ], [ IPI_FLAG, 'yellow' ], [ BICO_FLAG, 'yellow' ] ]:
            cell = table.cells_array[ 3, col ]
            if col == MONO_RG or cell.svar.get() == '1':
                fg = 'white' if color == 'red' else 'black'
                cell.widget.config( bg=color, fg=fg )



# coding: utf-8
"""
    EmgMixture.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import sys
import logging
import numpy as np
import scipy.stats as stats
from scipy.optimize import fsolve
from sklearn.cluster import KMeans
import molass_legacy.KekLib.DebugPlot as plt
from lmfit import minimize, Parameters
from molass_legacy.Models.ElutionCurveModels import emg

SQRT_2PI_INV = 1/np.sqrt(2*np.pi)

def emg_pdf(x, mu, sigma, tau=0, scale=1):
    return 1/sigma * SQRT_2PI_INV * emg(x, scale, mu, sigma, tau)

USE_MODE_AS_MU_INIT = True
USE_MODERATE_SIGMAS = False
MODERATE_SIGMAS_RATIO = 1.5
APPLY_PARAM_CONSTRAINTS = True

class EmgMixture:
    model_name = "EMG"

    def __init__(self, K, max_iter=100, random_state=None, anim_data=False):
        self.logger = logging.getLogger(__name__)
        self.K = K
        self.max_iter = max_iter
        self.random_state = random_state
        self.anim_data = anim_data

    def fit(self, X, bins=None):
        assert X.shape[1] == 1
        self.initilize(X, bins)

        for step in range(self.max_iter):
            self._e_step()
            self.update_params(step, self._m_step(step))

        print('result pi=', self.pi)
        print('result mu=', self.mu)
        print('result sigma=', self.sigma)
        print('result tau=', self.tau)

    def initilize(self, X, bins):
        self.N = X.shape[0]
        self.X = X
        self.X_  = X.flatten()
        self.bins = bins
        self.guess_initial_params()

    def guess_initial_params(self):
        X = self.X
        num_entire_points = X.shape[0]
        kmeans = KMeans(n_clusters= self.K, random_state=self.random_state)
        kmeans.fit(X)
        # predicted = kmeans.predict(X)
        predicted = kmeans.labels_
        labels = np.unique(predicted)
        init_params = []

        for label in labels:
            sy = X[predicted == label]

            M1 = np.mean(sy)
            if USE_MODE_AS_MU_INIT:
                # seems risky
                m, c = stats.mode(sy)
                mu = m[0,0]
            else:
                mu = M1

            num_points = sy.shape[0]
            pi = num_points/num_entire_points
            # M2 = pi*np.sum((sy - mu)**2)/num_points     # why pi* in Foley's implementation?
            M2 = np.sum((sy - M1)**2)/num_points
            sigma = np.sqrt(M2)
            # print([label], m, c, M2, np.sqrt(M2), sigma)
            init_params.append((pi, mu, sigma))

        sorted_params = np.array(sorted(init_params, key=lambda x:x[1]))

        self.pi = sorted_params[:,0]
        self.mu = sorted_params[:,1]
        self.sigma = sorted_params[:,2]
        self.tau = np.zeros(self.K)
        if False:
            print('init pi=', self.pi)
            print('init mu=', self.mu)
            print('init sigma=', self.sigma)
            print('init tau=', self.tau)
        if APPLY_PARAM_CONSTRAINTS:
            self.min_sigma = self.sigma*0.5
            self.max_sigma = self.sigma*2
            self.min_mu = np.max([np.zeros(self.K), self.mu - self.sigma], axis=0)
            self.max_mu = self.mu + self.sigma

        # assert np.sum(initial_pi) == 1
        assert abs(np.sum(self.pi) - 1) < 1e-10

        if False:
            plt.push()
            fig = plt.figure()
            ax = fig.gca()
            ax.hist(X.flatten(), bins=self.bins)
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            for mu in self.mu:
                ax.plot([mu, mu], [ymin, ymax], ':', color='red')
            plt.show()
            plt.pop()

        if self.anim_data:
            array_shape = (self.max_iter+1, self.K)
            self.pi_array = np.zeros(array_shape)
            self.mu_array = np.zeros(array_shape)
            self.sigma_array = np.zeros(array_shape)
            self.tau_array = np.zeros(array_shape)
            self.gamma_array = np.zeros((self.max_iter+1, self.N, self.K))
            self.set_anim_params(0)

    def set_anim_params(self, n):
        self.pi_array[n,:] = self.pi
        self.mu_array[n,:] = self.mu
        self.sigma_array[n,:] = self.sigma
        self.tau_array[n,:] = self.tau
        if n > 0:
            self.gamma_array[n,:,:] = self.gamma

    def _e_step(self):
        self.gamma = np.zeros((self.N, self.K))

        for k in range(self.K):
            # Posterior Distribution using Bayes Rule
            self.gamma[:,k] = self.pi[k] * emg_pdf(self.X_, self.mu[k], self.sigma[k], self.tau[k])

        # normalize across columns to make a valid probability
        # gamma_norm = np.sum(self.gamma, axis=1)[:,np.newaxis]
        gamma_norm = np.sum(self.gamma, axis=1)

        if False:
            print('pi=', self.pi)
            print('gamma_norm=', gamma_norm)
            plt.push()
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(gamma_norm.flatten())
            plt.show()
            plt.pop()

        positive = gamma_norm > 0
        for k in range(self.K):
            self.gamma[positive,k] /= gamma_norm[positive]      # normalize only where positive

        if False:
            from .GammaVisualizer import GammaVisualizer
            print(sys.getsizeof(self.gamma))
            vis = GammaVisualizer()
            vis.show(self.gamma)

    def _m_step(self, step, pi_only=False):

        # responsibilities for each gaussian
        self.pi = np.mean(self.gamma, axis=0)

        if pi_only:
            return

        W_ = np.sum(self.gamma, axis=0)[:,np.newaxis].T
        # print('GT.shape=', GT.shape)
        # print('GT[:,0:5]=', GT[:,0:5])
        # print('W_=', W_)

        """
        2.1.1 Central Moments and their Distribution Functions
        https://www.iue.tuwien.ac.at/phd/puchner/node17.html#SECTION00811000000000000000
        2.1.2 Probability Weighted Moments and their Distribution Functions
        https://www.iue.tuwien.ac.at/phd/puchner/node18.html

        Central moment - Wikipedia
        https://en.wikipedia.org/wiki/Central_moment
        """

        debug = False

        # M1 = (np.dot(self.gamma.T, self.X) / W_.T).T
        M1 = np.sum(self.gamma * self.X, axis=0) / W_
        M2 = np.sum(self.gamma * (self.X-M1)**2, axis=0) / W_
        M3 = np.sum(self.gamma * (self.X-M1)**3, axis=0) / W_

        if debug:
            print('M1=', M1)
            print('M2=', M2)
            print('M3=', M3)

        params = []
        for k in range(self.K):
            params.append(self.solve_params(k, M1[0,k], M2[0,k], M3[0,k]))

        params_ = np.array(params)
        if debug:
            print('params_=', params_)

        return params_

    def update_params(self, step, params_):
        self.mu = params_[:,0]
        if USE_MODERATE_SIGMAS:
            self.sigma = self.moderate_sigmas(params_[:,1])
        else:
            self.sigma = params_[:,1]
        self.tau = params_[:,2]

        if self.anim_data:
            self.set_anim_params(step+1)

    def moderate_sigmas(self, sigmas):
        if self.K > 1:
            L = list(range(self.K))
            k = np.argmax(sigmas)
            del L[k]
            sigmas_ = sigmas.copy()
            sigmas_[k] = min(sigmas[k], np.average(sigmas[L]) * MODERATE_SIGMAS_RATIO)
            return sigmas_
        else:
            return sigmas

    def solve_params(self, k, m1, m2, m3):
        last_mu = self.mu[k]
        last_sigma = self.sigma[k]
        last_tau = self.tau[k]

        """
            See 1998, N. Dyson, 
            for moment calculation
        """

        def equations(p):
            mu, sigma, tau = p
            # print('equations: ', mu, sigma, tau)
            return (
                        mu + tau - m1,
                        sigma**2 + tau**2 - m2,
                        2*tau**3 - m3,
                   )

        x, infodict, ier, mesg = fsolve(equations, (last_mu, last_sigma, last_tau), full_output=True)
        if ier in [1, 4, 5]:
            """
            5 : The iteration is not making good progress, as measured by the
                improvement from the last ten iterations.
            4 : The iteration is not making good progress, as measured by the
                improvement from the last five Jacobian evaluations.
            """
            mu, sigma, tau = x
        else:
            print('ier=', ier)
            self.logger.warning(mesg)
            self.fsolve_error = True
            mu, sigma, tau = last_mu, last_sigma, last_tau

        if APPLY_PARAM_CONSTRAINTS:
            """
            better constraints desired
            """
            mu = max(self.min_mu[k], min(self.max_mu[k], mu))
            sigma = max(self.min_sigma[k], min(self.max_sigma[k], sigma))
            ns = sigma*3
            tau = max(-ns, min(ns, tau))

        debug = False

        if debug:
            print([k], (last_mu, '->', mu))
            print([k], (last_sigma, '->', sigma))
            print([k], (last_tau, '->', tau))

        return mu, sigma, tau

    def get_anim_components(self, x, y, n):
        gy_list = []
        ty = np.zeros(len(x))
        for k in range(len(self.pi)):
            w = self.pi_array[n,k]
            m = self.mu_array[n,k]
            s = self.sigma_array[n,k]
            t = self.tau_array[n,k]
            # print([n,k], w, m, s, t)
            gy = emg_pdf(x, m, s, t, w)
            gy_list.append([gy, m])
            ty += gy

        sorted_gy_list = sorted(gy_list, key=lambda x:x[1])

        def obj_func(params):
            S   = params['S']
            return ty*S - y

        params = Parameters()
        S_init = np.max(y)/np.max(ty)
        params.add('S', value=S_init, min=0, max=S_init*100 )
        result = minimize(obj_func, params, args=())

        scale = result.params['S'].value
        return [ ty*scale ] + [ scale*rec[0] for rec in sorted_gy_list ]

    def get_anim_C(self, x, y, n, total=False):
        gy_list = self.get_anim_components(x, y, n)
        if total:
            return np.array(gy_list[1:]), gy_list[0]
        else:
            return np.array(gy_list[1:])

    def get_peak_mean_x(self):
        return self.mu

def get_sorted_params(eghmm):
    # sort results in the order of the means
    params = zip(eghmm.mu, eghmm.sigma, eghmm.tau, eghmm.pi)
    sorted_params = sorted([ (m, s, t, w) for m, s, t, w in params], key=lambda p:p[0])
    return sorted_params

def get_curves(eghmm, x):
    sorted_params = get_sorted_params(eghmm)

    # plot the results
    gy_list = []
    ty = np.zeros(len(x))
    for m, s, t, w in sorted_params:
        gy = emg_pdf(x, m, s, t, w)
        gy_list.append(gy)
        ty += gy

    return ty, gy_list

